# Author: Idan Gutman
# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import bs4_parser
from sickrage.providers import TorrentProvider


class SceneTimeProvider(TorrentProvider):
    def __init__(self):
        super(SceneTimeProvider, self).__init__("SceneTime", 'https://www.scenetime.com', True)

        self.urls.update({
            'login': '{base_url}/takelogin.php'.format(**self.urls),
            'detail': '{base_url}/details.php?id=%s'.format(**self.urls),
            'search': '{base_url}/browse_API.php'.format(**self.urls),
            'download': '{base_url}/download.php/%s/%s'.format(**self.urls)
        })

        self.username = None
        self.password = None

        self.minseed = None
        self.minleech = None

        self.enable_cookies = True
        self.required_cookies = ('uid', 'pass')

        self.categories = [2, 42, 9, 63, 77, 79, 100, 83]

        self.cache = TVCache(self, min_time=20)

    def login(self):
        return self.cookie_login('sign in')

    def search(self, search_params, age=0, ep_obj=None):
        results = []

        if not self.login():
            return results

        for mode in search_params.keys():
            sickrage.app.log.debug("Search Mode: %s" % mode)
            for search_string in search_params[mode]:
                if mode != 'RSS':
                    sickrage.app.log.debug("Search string: %s " % search_string)

                query = {'sec': 'jax', 'cata': 'yes', 'search': search_string}
                query.update({"c%s" % i: 1 for i in self.categories})

                try:
                    data = self.session.post(self.urls['search'], data=query).text
                    results += self.parse(data, mode)
                except Exception:
                    sickrage.app.log.debug("No data returned from provider")

        return results

    def parse(self, data, mode):
        """
        Parse search results from data
        :param data: response data
        :param mode: search mode
        :return: search results
        """

        results = []

        with bs4_parser(data) as html:
            torrent_rows = html.findAll('tr')

            # Continue only if one Release is found
            if len(torrent_rows) < 2:
                sickrage.app.log.debug("Data returned from provider does not contain any torrents")
                return results

            # Scenetime apparently uses different number of cells in #torrenttable based
            # on who you are. This works around that by extracting labels from the first
            # <tr> and using their index to find the correct download/seeders/leechers td.
            labels = [label.get_text() for label in torrent_rows[0].find_all('td')]

            for result in torrent_rows[1:]:
                cells = result.find_all('td')

                link = cells[labels.index('Name')].find('a')

                full_id = link['href'].replace('details.php?id=', '')
                torrent_id = full_id.split("&")[0]

                try:
                    title = link.contents[0].get_text()
                    filename = "%s.torrent" % title.replace(" ", ".")
                    download_url = self.urls['download'] % (torrent_id, filename)

                    int(cells[labels.index('Seeders')].get_text())
                    seeders = int(cells[labels.index('Seeders')].get_text())
                    leechers = int(cells[labels.index('Leechers')].get_text())
                    # FIXME
                    size = -1

                    if not all([title, download_url]):
                        continue

                    results += [
                        {'title': title, 'link': download_url, 'size': size, 'seeders': seeders, 'leechers': leechers}
                    ]

                    if mode != 'RSS':
                        sickrage.app.log.debug("Found result: {}".format(title))
                except Exception:
                    sickrage.app.log.error("Failed parsing provider")

        return results



from __future__ import absolute_import, division, print_function

import functools
import logging
import warnings

from trakt.core.errors import log_request_error
from trakt.core.exceptions import RequestFailedError, ServerError, ClientError
from trakt.core.helpers import try_convert
from trakt.core.pagination import PaginationIterator
from trakt.helpers import setdefault

log = logging.getLogger(__name__)


def authenticated(func):
    @functools.wraps(func)
    def wrap(*args, **kwargs):
        if 'authenticated' not in kwargs:
            kwargs['authenticated'] = True

        return func(*args, **kwargs)

    return wrap


def application(func):
    @functools.wraps(func)
    def wrap(*args, **kwargs):
        if args and isinstance(args[0], Interface):
            interface = args[0]

            setdefault(kwargs, {
                'app_version': interface.client.configuration['app.version'],
                'app_date': interface.client.configuration['app.date']
            }, lambda key, value: value)

        return func(*args, **kwargs)

    return wrap


class Interface(object):
    path = None

    def __init__(self, client):
        self.client = client

    def __getitem__(self, name):
        if hasattr(self, name):
            return getattr(self, name)

        raise ValueError('Unknown action "%s" on %s' % (name, self))

    @property
    def http(self):
        if not self.client:
            return None

        return self.client.http.configure(self.path)

    def get_data(self, response, exceptions=False, pagination=False, parse=True):
        if response is None:
            if exceptions:
                raise RequestFailedError('No response available')

            log.warn('Request failed (no response returned)')
            return None

        # Return response, if parse=False
        if not parse:
            return response

        # Check status code, log any errors
        error = False

        if response.status_code < 200 or response.status_code >= 300:
            log_request_error(log, response)

            # Raise an exception (if enabled)
            if exceptions:
                if response.status_code >= 500:
                    raise ServerError(response)
                else:
                    raise ClientError(response)

            # Set error flag
            error = True

        # Return `None` if we encountered an error, return response data
        if error:
            return None

        # Check for pagination response
        page_count = try_convert(response.headers.get('x-pagination-page-count'), int)

        if page_count and page_count > 1:
            if pagination:
                return PaginationIterator(self.client, response)

            warnings.warn(
                'Unhandled pagination response, more pages can be returned with `pagination=True`',
                stacklevel=3
            )

        # Parse response, return data
        content_type = response.headers.get('content-type')

        if content_type and content_type.startswith('application/json'):
            # Try parse json response
            try:
                data = response.json()
            except Exception as e:
                log.warning('unable to parse JSON response: %s', e)
                return None
        else:
            log.debug('response returned content-type: %r, falling back to raw data', content_type)

            # Fallback to raw content
            data = response.content

        return data


class InterfaceProxy(object):
    def __init__(self, interface, args):
        self.interface = interface
        self.args = list(args)

    def __getattr__(self, name):
        value = getattr(self.interface, name)

        if not callable(value):
            return value

        @functools.wraps(value)
        def wrap(*args, **kwargs):
            args = self.args + list(args)

            return value(*args, **kwargs)

        return wrap



from __future__ import absolute_import, division, print_function

import calendar
import logging
import time
from datetime import datetime, timedelta
from threading import Thread

import requests
from trakt.core.emitter import Emitter
from trakt.interfaces.base import Interface

log = logging.getLogger(__name__)


class DeviceOAuthInterface(Interface):
    path = 'oauth/device'

    def code(self, **kwargs):
        client_id = self.client.configuration['client.id']

        if not client_id:
            raise ValueError('"client.id" configuration parameter is required')

        response = self.http.post(
            'code',
            data={
                'client_id': client_id
            }
        )

        data = self.get_data(response, **kwargs)

        if isinstance(data, requests.Response):
            return data

        if not data:
            return None

        return data

    def poll(self, device_code, expires_in, interval, **kwargs):
        """Construct the device authentication poller.

        :param device_code: Device authentication code
        :type device_code: str

        :param expires_in: Device authentication code expiry (in seconds)
        :type         in: int

        :param interval: Device authentication poll interval
        :type interval: int

        :rtype: DeviceOAuthPoller
        """
        return DeviceOAuthPoller(self.client, device_code, expires_in, interval)

    def token(self, device_code, **kwargs):
        client_id = self.client.configuration['client.id']
        client_secret = self.client.configuration['client.secret']

        if not client_id:
            raise ValueError('"client.id" and "client.secret" configuration parameters are required')

        response = self.http.post(
            'token',
            data={
                'client_id': client_id,
                'client_secret': client_secret,

                'code': device_code
            }
        )

        data = self.get_data(response, **kwargs)

        if isinstance(data, requests.Response):
            return data

        if not data:
            return None

        return data


class DeviceOAuthPoller(Interface, Emitter):
    def __init__(self, client, device_code, expires_in, interval):
        super(DeviceOAuthPoller, self).__init__(client)

        self.device_code = device_code
        self.expires_in = expires_in
        self.interval = interval

        # Calculate code expiry date/time
        self.expires_at = datetime.utcnow() + timedelta(seconds=self.expires_in)

        # Private attributes
        self._abort = False
        self._active = False
        self._running = False
        self._thread = None

    @property
    def active(self):
        return self._active

    def has_expired(self):
        return datetime.utcnow() > self.expires_at

    def start(self, daemon=None):
        if self._active or self._thread:
            raise Exception('Poller already started')

        # Construct thread process wrapper
        def wrapper():
            try:
                self._process()
            except Exception as ex:
                log.warn('Exception raised in DeviceOAuthPoller: %s', ex, exc_info=True)
            finally:
                self._active = False
                self._running = False

                if self._abort:
                    self.emit('aborted')

        # Construct poller thread
        self._thread = Thread(
            target=wrapper,
            name='%s:%s' % (DeviceOAuthPoller.__module__, DeviceOAuthPoller.__name__)
        )

        # Set `daemon` state
        if daemon is not None:
            self._thread.daemon = daemon

        # Start polling
        self._abort = False
        self._active = True
        self._running = True
        self._thread.start()

    def stop(self):
        # Flag as thread abort
        self._abort = True

        # Flag thread to stop
        self._running = False

    def _process(self):
        while self._running:
            # Ensure code hasn't expired yet
            if self.has_expired():
                self.emit('expired')
                break

            # Trigger "poll" event, check if we should continue polling
            if not self._should_poll():
                self.stop()
                break

            # Poll for token
            response = self.client['oauth/device'].token(self.device_code, parse=False)

            if response:
                # Parse authorization
                data = self.get_data(response)

                if 'created_at' not in data:
                    data['created_at'] = calendar.timegm(datetime.utcnow().utctimetuple())

                # Authentication complete
                self.emit('authenticated', data)
                break

            # Sleep for defined interval
            time.sleep(self.interval)

    def _poll_callback(self, state=True):
        self._abort = not state

    def _should_poll(self):
        # Assume poller should abort if `callback` isn't fired
        self._abort = True

        # Trigger "poll" event
        self.emit('poll', self._poll_callback)

        # Continue polling if `abort` flag isn't set
        return not self._abort



from __future__ import absolute_import, division, print_function

from trakt.helpers import build_url
from trakt.interfaces.base import Interface


class PinOAuthInterface(Interface):
    path = 'oauth/pin'

    def url(self):
        app_id = self.client.configuration['app.id']

        if not app_id:
            raise ValueError('"app.id" configuration parameter is required to generate the PIN authentication url')

        return build_url(
            self.client.site_url,
            'pin', app_id
        )



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import os
import re
import shutil
import tarfile
import time
import traceback
from sqlite3 import OperationalError

from CodernityDB.database_super_thread_safe import SuperThreadSafeDatabase
from CodernityDB.index import IndexNotFoundException, IndexConflict, IndexException
from CodernityDB.storage import IU_Storage

import sickrage
from sickrage.core.helpers import randomString


def Custom_IU_Storage_get(self, start, size, status='c'):
    if status == 'd':
        return None
    else:
        self._f.seek(start)
        return self.data_from(self._f.read(size))


class srDatabase(object):
    _indexes = {}
    _migrate_list = {}

    def __init__(self, name=''):
        self.name = name
        self.old_db_path = ''

        self.db_path = os.path.join(sickrage.app.data_dir, 'database', self.name)
        self.db = SuperThreadSafeDatabase(self.db_path)

    def initialize(self):
        # Remove database folder if both exists
        if self.db.exists() and os.path.isfile(self.old_db_path):
            self.db.open()
            self.db.destroy()

        if self.db.exists():
            # Backup before start and cleanup old backups
            backup_path = os.path.join(sickrage.app.data_dir, 'db_backup', self.name)
            backup_count = 5
            existing_backups = []
            if not os.path.isdir(backup_path): os.makedirs(backup_path)

            for root, dirs, files in os.walk(backup_path):
                # Only consider files being a direct child of the backup_path
                if root == backup_path:
                    for backup_file in sorted(files):
                        ints = re.findall('\d+', backup_file)

                        # Delete non zip files
                        if len(ints) != 1:
                            try:
                                os.remove(os.path.join(root, backup_file))
                            except:
                                pass
                        else:
                            existing_backups.append((int(ints[0]), backup_file))
                else:
                    # Delete stray directories.
                    shutil.rmtree(root)

            # Remove all but the last 5
            for eb in existing_backups[:-backup_count]:
                os.remove(os.path.join(backup_path, eb[1]))

            # Create new backup
            new_backup = os.path.join(backup_path, '%s.tar.gz' % int(time.time()))
            with tarfile.open(new_backup, 'w:gz') as zipf:
                for root, dirs, files in os.walk(self.db_path):
                    for zfilename in files:
                        zipf.add(os.path.join(root, zfilename),
                                 arcname='database/%s/%s' % (
                                     self.name,
                                     os.path.join(root[len(self.db_path) + 1:], zfilename))
                                 )

            self.db.open()
        else:
            self.db.create()

        # setup database indexes
        self.setupIndexes()

    def compact(self, try_repair=True, **kwargs):
        # Removing left over compact files
        for f in os.listdir(self.db.path):
            for x in ['_compact_buck', '_compact_stor']:
                if f[-len(x):] == x:
                    os.unlink(os.path.join(self.db.path, f))

        try:
            start = time.time()
            size = float(self.db.get_db_details().get('size', 0))
            sickrage.app.log.info(
                'Compacting {} database, current size: {}MB'.format(self.name, round(size / 1048576, 2)))

            self.db.compact()

            new_size = float(self.db.get_db_details().get('size', 0))
            sickrage.app.log.info(
                'Done compacting {} database in {}s, new size: {}MB, saved: {}MB'.format(
                    self.name, round(time.time() - start, 2),
                    round(new_size / 1048576, 2), round((size - new_size) / 1048576, 2))
            )
        except (IndexException, AttributeError, TypeError) as e:
            if try_repair:
                sickrage.app.log.debug('Something wrong with indexes, trying repair')

                # Remove all indexes
                old_indexes = self._indexes.keys()
                for index_name in old_indexes:
                    try:
                        self.db.destroy_index(index_name)
                    except IndexNotFoundException:
                        pass
                    except:
                        sickrage.app.log.debug('Failed removing old index %s', index_name)

                # Add them again
                for index_name in self._indexes:
                    try:
                        self.db.add_index(self._indexes[index_name](self.db.path, index_name))
                        self.db.reindex_index(index_name)
                    except IndexConflict:
                        pass
                    except:
                        sickrage.app.log.debug('Failed adding index %s', index_name)
                        raise

                self.compact(try_repair=False)
            else:
                sickrage.app.log.debug('Failed compact: {}'.format(traceback.format_exc()))
        except:
            sickrage.app.log.debug('Failed compact: {}'.format(traceback.format_exc()))

    def setupIndexes(self):
        # setup database indexes
        for index_name in self._indexes:
            try:
                # Make sure store and bucket don't exist
                exists = []
                for x in ['buck', 'stor']:
                    full_path = os.path.join(self.db.path, '%s_%s' % (index_name, x))
                    if os.path.exists(full_path):
                        exists.append(full_path)

                if index_name not in self.db.indexes_names:
                    # Remove existing buckets if index isn't there
                    for x in exists:
                        os.unlink(x)

                    self.db.add_index(self._indexes[index_name](self.db.path, index_name))
                    self.db.reindex_index(index_name)
                else:
                    # Previous info
                    previous_version = self.db.indexes_names[index_name]._version
                    current_version = self._indexes[index_name]._version

                    self.check_versions(index_name, current_version, previous_version)
            except:
                sickrage.app.log.debug('Failed adding index {}'.format(index_name))

    def check_versions(self, index_name, current_version, previous_version):
        # Only edit index if versions are different
        if previous_version < current_version:
            self.db.destroy_index(self.db.indexes_names[index_name])
            self.db.add_index(self._indexes[index_name](self.db.path, index_name))
            self.db.reindex_index(index_name)

    def close(self):
        self.db.close()

    def cleanup(self):
        pass

    @property
    def opened(self):
        return self.db.opened

    def check_integrity(self):
        for index_name in self._indexes:
            try:
                for x in self.db.all(index_name):
                    try:
                        self.get('id', x.get('_id'))
                    except (ValueError, TypeError) as e:
                        self.delete(self.get(index_name, x.get('key')))
            except Exception as e:
                if index_name in self.db.indexes_names:
                    self.db.destroy_index(self.db.indexes_names[index_name])

    def migrate(self):
        if os.path.isfile(self.old_db_path):
            sickrage.app.log.info('=' * 30)
            sickrage.app.log.info('Migrating %s database, please wait...', self.name)
            migrate_start = time.time()

            import sqlite3
            conn = sqlite3.connect(self.old_db_path)
            conn.text_factory = lambda x: (x.decode('utf-8', 'ignore'))

            migrate_data = {}
            rename_old = False

            try:
                c = conn.cursor()

                for ml in self._migrate_list:
                    migrate_data[ml] = {}
                    rows = self._migrate_list[ml]

                    try:
                        c.execute('SELECT {} FROM `{}`'.format('`' + '`,`'.join(rows) + '`', ml))
                    except:
                        # ignore faulty destination_id database
                        rename_old = True
                        raise

                    for p in c.fetchall():
                        columns = {}
                        for row in self._migrate_list[ml]:
                            columns[row] = p[rows.index(row)]

                        if not migrate_data[ml].get(p[0]):
                            migrate_data[ml][p[0]] = columns
                        else:
                            if not isinstance(migrate_data[ml][p[0]], list):
                                migrate_data[ml][p[0]] = [migrate_data[ml][p[0]]]
                            migrate_data[ml][p[0]].append(columns)

                sickrage.app.log.info('Getting data took %s', (time.time() - migrate_start))

                if not self.db.opened:
                    return

                for t_name in migrate_data:
                    t_data = migrate_data.get(t_name, {})
                    sickrage.app.log.info('Importing %s %s' % (len(t_data), t_name))
                    for k, v in t_data.items():
                        if isinstance(v, list):
                            for d in v:
                                d.update({'_t': t_name})
                                self.insert(d)
                        else:
                            v.update({'_t': t_name})
                            self.insert(v)

                sickrage.app.log.info('Total migration took %s', (time.time() - migrate_start))
                sickrage.app.log.info('=' * 30)

                rename_old = True
            except OperationalError:
                sickrage.app.log.debug('Migrating from unsupported/corrupt %s database version', self.name)
                rename_old = True
            except:
                sickrage.app.log.debug('Migration of %s database failed', self.name)
            finally:
                conn.close()

            # rename old database
            if rename_old:
                random = randomString()
                sickrage.app.log.info('Renaming old database to %s.%s_old' % (self.old_db_path, random))
                os.rename(self.old_db_path, '{}.{}_old'.format(self.old_db_path, random))

                if os.path.isfile(self.old_db_path + '-wal'):
                    os.rename(self.old_db_path + '-wal', '{}-wal.{}_old'.format(self.old_db_path, random))
                if os.path.isfile(self.old_db_path + '-shm'):
                    os.rename(self.old_db_path + '-shm', '{}-shm.{}_old'.format(self.old_db_path, random))

    def all(self, *args, **kwargs):
        kwargs['with_doc'] = True
        return (x['doc'] for x in self.db.all(*args, **kwargs))

    def get_many(self, *args, **kwargs):
        kwargs['with_doc'] = True
        return (x['doc'] for x in self.db.get_many(*args, **kwargs))

    def get(self, *args, **kwargs):
        kwargs['with_doc'] = True
        data = self.db.get(*args, **kwargs)
        return data.get('doc', data)

    def delete(self, *args):
        return self.db.delete(*args)

    def update(self, *args):
        return self.db.update(*args)

    def insert(self, *args):
        return self.db.insert(*args)

# Monkey-Patch storage to suppress logging messages
IU_Storage.get = Custom_IU_Storage_get



# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import datetime
from xml.etree.ElementTree import Element, ElementTree, SubElement

import sickrage
from sickrage.core.common import dateFormat
from sickrage.core.exceptions import ShowNotFoundException
from sickrage.core.helpers import indentXML
from sickrage.indexers import IndexerApi
from sickrage.indexers.exceptions import indexer_episodenotfound, \
    indexer_error, indexer_seasonnotfound, indexer_shownotfound
from sickrage.metadata import GenericMetadata


class KODI_12PlusMetadata(GenericMetadata):
    """
    Metadata generation class for KODI 12+.

    The following file structure is used:

    show_root/tvshow.nfo                    (show metadata)
    show_root/fanart.jpg                    (fanart)
    show_root/poster.jpg                    (poster)
    show_root/banner.jpg                    (banner)
    show_root/Season ##/filename.ext        (*)
    show_root/Season ##/filename.nfo        (episode metadata)
    show_root/Season ##/filename-thumb.jpg  (episode thumb)
    show_root/season##-poster.jpg           (season posters)
    show_root/season##-banner.jpg           (season banners)
    show_root/season-all-poster.jpg         (season all poster)
    show_root/season-all-banner.jpg         (season all banner)
    """

    def __init__(self,
                 show_metadata=False,
                 episode_metadata=False,
                 fanart=False,
                 poster=False,
                 banner=False,
                 episode_thumbnails=False,
                 season_posters=False,
                 season_banners=False,
                 season_all_poster=False,
                 season_all_banner=False):

        GenericMetadata.__init__(self,
                                 show_metadata,
                                 episode_metadata,
                                 fanart,
                                 poster,
                                 banner,
                                 episode_thumbnails,
                                 season_posters,
                                 season_banners,
                                 season_all_poster,
                                 season_all_banner)

        self.name = 'KODI 12+'

        self.poster_name = "poster.jpg"
        self.season_all_poster_name = "season-all-poster.jpg"

        # web-ui metadata template
        self.eg_show_metadata = "tvshow.nfo"
        self.eg_episode_metadata = "Season##\\<i>filename</i>.nfo"
        self.eg_fanart = "fanart.jpg"
        self.eg_poster = "poster.jpg"
        self.eg_banner = "banner.jpg"
        self.eg_episode_thumbnails = "Season##\\<i>filename</i>-thumb.jpg"
        self.eg_season_posters = "season##-poster.jpg"
        self.eg_season_banners = "season##-banner.jpg"
        self.eg_season_all_poster = "season-all-poster.jpg"
        self.eg_season_all_banner = "season-all-banner.jpg"

    def _show_data(self, show_obj):
        """
        Creates an elementTree XML structure for an KODI-style tvshow.nfo and
        returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """

        show_ID = show_obj.indexerid

        indexer_lang = show_obj.lang or sickrage.app.config.indexer_default_language
        lINDEXER_API_PARMS = IndexerApi(show_obj.indexer).api_params.copy()

        lINDEXER_API_PARMS['language'] = indexer_lang

        if show_obj.dvdorder != 0:
            lINDEXER_API_PARMS['dvdorder'] = True

        t = IndexerApi(show_obj.indexer).indexer(**lINDEXER_API_PARMS)

        tv_node = Element("tvshow")

        try:
            myShow = t[int(show_ID)]
        except indexer_shownotfound:
            sickrage.app.log.error("Unable to find show with id " + str(show_ID) + " on " + IndexerApi(
                show_obj.indexer).name + ", skipping it")
            raise

        except indexer_error:
            sickrage.app.log.error(
                "" + IndexerApi(show_obj.indexer).name + " is down, can't use its data to add this show")
            raise

        # check for title and id
        if not (getattr(myShow, 'seriesname', None) and getattr(myShow, 'id', None)):
            sickrage.app.log.info("Incomplete info for show with id " + str(show_ID) + " on " + IndexerApi(
                show_obj.indexer).name + ", skipping it")
            return False

        title = SubElement(tv_node, "title")
        title.text = myShow["seriesname"]

        if getattr(myShow, 'rating', None):
            rating = SubElement(tv_node, "rating")
            rating.text = myShow["rating"]

        if getattr(myShow, 'firstaired', None):
            try:
                year_text = str(datetime.datetime.strptime(myShow["firstaired"], dateFormat).year)
                if year_text:
                    year = SubElement(tv_node, "year")
                    year.text = year_text
            except:
                pass

        if getattr(myShow, 'overview', None):
            plot = SubElement(tv_node, "plot")
            plot.text = myShow["overview"]

        # if getattr(myShow, 'id', None):
        #    episodeguide = SubElement(tv_node, "episodeguide")
        #    episodeguideurl = SubElement(episodeguide, "url")
        #    episodeguideurl.text = IndexerApi(show_obj.indexer).config['base_url'] + str(
        #        myShow["id"]) + '/all/en.zip'

        if getattr(myShow, 'contentrating', None):
            mpaa = SubElement(tv_node, "mpaa")
            mpaa.text = myShow["contentrating"]

        if getattr(myShow, 'id', None):
            indexerid = SubElement(tv_node, "id")
            indexerid.text = str(myShow["id"])

        if getattr(myShow, 'genre', None) and isinstance(myShow["genre"], basestring):
            genre = SubElement(tv_node, "genre")
            genre.text = " / ".join(x.strip() for x in myShow["genre"].split('|') if x.strip())

        if getattr(myShow, 'firstaired', None):
            premiered = SubElement(tv_node, "premiered")
            premiered.text = myShow["firstaired"]

        if getattr(myShow, 'network', None):
            studio = SubElement(tv_node, "studio")
            studio.text = myShow["network"].strip()

        for actor in t.actors(int(show_ID)):
            cur_actor = SubElement(tv_node, "actor")

            if 'name' in actor and actor['name'].strip():
                cur_actor_name = SubElement(cur_actor, "name")
                cur_actor_name.text = actor['name'].strip()
            else:
                continue

            if 'role' in actor and actor['role'].strip():
                cur_actor_role = SubElement(cur_actor, "role")
                cur_actor_role.text = actor['role'].strip()

            if 'image' in actor and actor['image'].strip():
                cur_actor_thumb = SubElement(cur_actor, "thumb")
                cur_actor_thumb.text = actor['image'].strip()

        # Make it purdy
        indentXML(tv_node)

        data = ElementTree(tv_node)

        return data

    def _ep_data(self, ep_obj):
        """
        Creates an elementTree XML structure for an KODI-style episode.nfo and
        returns the resulting data object.
            show_obj: a TVEpisode instance to create the NFO for
        """

        eps_to_write = [ep_obj] + ep_obj.relatedEps

        indexer_lang = ep_obj.show.lang or sickrage.app.config.indexer_default_language

        lINDEXER_API_PARMS = IndexerApi(ep_obj.show.indexer).api_params.copy()

        lINDEXER_API_PARMS['language'] = indexer_lang

        if ep_obj.show.dvdorder != 0:
            lINDEXER_API_PARMS['dvdorder'] = True

        try:
            t = IndexerApi(ep_obj.show.indexer).indexer(**lINDEXER_API_PARMS)
            myShow = t[ep_obj.show.indexerid]
        except indexer_shownotfound as e:
            raise ShowNotFoundException(e.message)
        except indexer_error as e:
            sickrage.app.log.error("Unable to connect to {} while creating meta files - skipping - {}".format(
                IndexerApi(
                    ep_obj.show.indexer).name, e))
            return

        if len(eps_to_write) > 1:
            rootNode = Element("kodimultiepisode")
        else:
            rootNode = Element("episodedetails")

        # write an NFO containing info for all matching episodes
        for curEpToWrite in eps_to_write:

            try:
                myEp = myShow[curEpToWrite.season][curEpToWrite.episode]
            except (indexer_episodenotfound, indexer_seasonnotfound):
                sickrage.app.log.info(
                    "Unable to find episode %dx%d on %s... has it been removed? Should I delete from db?" %
                    (curEpToWrite.season, curEpToWrite.episode, IndexerApi(ep_obj.show.indexer).name))
                return None

            if not getattr(myEp, 'firstaired', None):
                myEp["firstaired"] = str(datetime.date.fromordinal(1))

            if not getattr(myEp, 'episodename', None):
                sickrage.app.log.debug("Not generating nfo because the ep has no title")
                return None

            sickrage.app.log.debug(
                "Creating metadata for episode " + str(ep_obj.season) + "x" + str(ep_obj.episode))

            if len(eps_to_write) > 1:
                episode = SubElement(rootNode, "episodedetails")
            else:
                episode = rootNode

            if getattr(myEp, 'episodename', None):
                title = SubElement(episode, "title")
                title.text = myEp['episodename']

            if getattr(myShow, 'seriesname', None):
                showtitle = SubElement(episode, "showtitle")
                showtitle.text = myShow['seriesname']

            season = SubElement(episode, "season")
            season.text = str(curEpToWrite.season)

            episodenum = SubElement(episode, "episode")
            episodenum.text = str(curEpToWrite.episode)

            uniqueid = SubElement(episode, "uniqueid")
            uniqueid.text = str(curEpToWrite.indexerid)

            if curEpToWrite.airdate != datetime.date.fromordinal(1):
                aired = SubElement(episode, "aired")
                aired.text = str(curEpToWrite.airdate)

            if getattr(myEp, 'overview', None):
                plot = SubElement(episode, "plot")
                plot.text = myEp['overview']

            if curEpToWrite.season and getattr(myShow, 'runtime', None):
                runtime = SubElement(episode, "runtime")
                runtime.text = myShow["runtime"]

            if getattr(myEp, 'airsbefore_season', None):
                displayseason = SubElement(episode, "displayseason")
                displayseason.text = myEp['airsbefore_season']

            if getattr(myEp, 'airsbefore_episode', None):
                displayepisode = SubElement(episode, "displayepisode")
                displayepisode.text = myEp['airsbefore_episode']

            if getattr(myEp, 'filename', None):
                thumb = SubElement(episode, "thumb")
                thumb.text = myEp['filename'].strip()

            # watched = SubElement(episode, "watched")
            # watched.text = 'false'

            if getattr(myEp, 'writer', None):
                ep_credits = SubElement(episode, "credits")
                ep_credits.text = myEp['writer'].strip()

            if getattr(myEp, 'director', None):
                director = SubElement(episode, "director")
                director.text = myEp['director'].strip()

            if getattr(myEp, 'rating', None):
                rating = SubElement(episode, "rating")
                rating.text = myEp['rating']

            if getattr(myEp, 'gueststars', None) and isinstance(myEp['gueststars'], basestring):
                for actor in (x.strip() for x in myEp['gueststars'].split('|') if x.strip()):
                    cur_actor = SubElement(episode, "actor")
                    cur_actor_name = SubElement(cur_actor, "name")
                    cur_actor_name.text = actor

            for actor in t.actors(int(ep_obj.show.indexerid)):
                cur_actor = SubElement(episode, "actor")

                if 'name' in actor and actor['name'].strip():
                    cur_actor_name = SubElement(cur_actor, "name")
                    cur_actor_name.text = actor['name'].strip()
                else:
                    continue

                if 'role' in actor and actor['role'].strip():
                    cur_actor_role = SubElement(cur_actor, "role")
                    cur_actor_role.text = actor['role'].strip()

                if 'image' in actor and actor['image'].strip():
                    cur_actor_thumb = SubElement(cur_actor, "thumb")
                    cur_actor_thumb.text = actor['image'].strip()

        # Make it purdy
        indentXML(rootNode)
        data = ElementTree(rootNode)
        return data



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import datetime
import time

import feedparser
from CodernityDB.database import RecordNotFound

import sickrage
from sickrage.core.common import Quality
from sickrage.core.exceptions import AuthException, EpisodeNotFoundException
from sickrage.core.helpers import findCertainShow, show_names, validate_url, is_ip_private
from sickrage.core.nameparser import InvalidNameException, NameParser, InvalidShowException
from sickrage.core.websession import WebSession


class TVCache(object):
    def __init__(self, provider, min_time=10, search_params=None):
        self.provider = provider
        self.providerID = self.provider.id
        self.min_time = min_time
        self.search_params = search_params or {'RSS': ['']}

    def clear(self):
        if self.shouldClearCache():
            [sickrage.app.cache_db.delete(x) for x in
             sickrage.app.cache_db.get_many('providers', self.providerID)]

    def _get_title_and_url(self, item):
        return self.provider._get_title_and_url(item)

    def _get_result_stats(self, item):
        return self.provider._get_result_stats(item)

    def _get_size(self, item):
        return self.provider._get_size(item)

    def _get_rss_data(self):
        if self.search_params:
            return {'entries': self.provider.search(self.search_params)}

    def _check_auth(self, data):
        return True

    def check_item(self, title, url):
        return True

    def update(self):
        # check if we should update
        if self.should_update():
            try:
                data = self._get_rss_data()
                if not self._check_auth(data):
                    return False

                # clear cache
                self.clear()

                # set updated
                self.last_update = datetime.datetime.today()

                [self._parseItem(item) for item in data['entries']]
            except AuthException as e:
                sickrage.app.log.warning("Authentication error: {}".format(e.message))
                return False
            except Exception as e:
                sickrage.app.log.debug(
                    "Error while searching {}, skipping: {}".format(self.provider.name, repr(e)))
                return False

        return True

    def getRSSFeed(self, url, params=None):
        try:
            if self.provider.login():
                resp = WebSession().get(url, params=params).text
                return feedparser.parse(resp)
        except Exception as e:
            sickrage.app.log.debug("RSS Error: {}".format(e.message))

        return feedparser.FeedParserDict()

    def _translateTitle(self, title):
        return '' + title.replace(' ', '.')

    def _translateLinkURL(self, url):
        return url.replace('&amp;', '&')

    def _parseItem(self, item):
        title, url = self._get_title_and_url(item)
        seeders, leechers = self._get_result_stats(item)
        size = self._get_size(item)

        self.check_item(title, url)

        if title and url:
            self.addCacheEntry(self._translateTitle(title), self._translateLinkURL(url), seeders, leechers, size)
        else:
            sickrage.app.log.debug(
                "The data returned from the " + self.provider.name + " feed is incomplete, this result is unusable")

    @property
    def last_update(self):
        try:
            dbData = sickrage.app.cache_db.get('lastUpdate', self.providerID)
            lastTime = int(dbData["time"])
            if lastTime > int(time.mktime(datetime.datetime.today().timetuple())): lastTime = 0
        except RecordNotFound:
            lastTime = 0

        return datetime.datetime.fromtimestamp(lastTime)

    @last_update.setter
    def last_update(self, toDate):
        try:
            dbData = sickrage.app.cache_db.get('lastUpdate', self.providerID)
            dbData['time'] = int(time.mktime(toDate.timetuple()))
            sickrage.app.cache_db.update(dbData)
        except RecordNotFound:
            sickrage.app.cache_db.insert({
                '_t': 'lastUpdate',
                'provider': self.providerID,
                'time': int(time.mktime(toDate.timetuple()))
            })

    @property
    def last_search(self):
        try:
            dbData = sickrage.app.cache_db.get('lastSearch', self.providerID)
            lastTime = int(dbData["time"])
            if lastTime > int(time.mktime(datetime.datetime.today().timetuple())): lastTime = 0
        except RecordNotFound:
            lastTime = 0

        return datetime.datetime.fromtimestamp(lastTime)

    @last_search.setter
    def last_search(self, toDate):
        try:
            dbData = sickrage.app.cache_db.get('lastSearch', self.providerID)
            dbData['time'] = int(time.mktime(toDate.timetuple()))
            sickrage.app.cache_db.update(dbData)
        except RecordNotFound:
            sickrage.app.cache_db.insert({
                '_t': 'lastUpdate',
                'provider': self.providerID,
                'time': int(time.mktime(toDate.timetuple()))
            })

    def should_update(self):
        if sickrage.app.developer: return True

        # if we've updated recently then skip the update
        if datetime.datetime.today() - self.last_update < datetime.timedelta(minutes=self.min_time):
            return False
        return True

    def shouldClearCache(self):
        # if daily search hasn't used our previous results yet then don't clear the cache
        if self.last_update > self.last_search:
            return False
        return True

    def addCacheEntry(self, name, url, seeders, leechers, size):
        # check for existing entry in cache
        if len([x for x in sickrage.app.cache_db.get_many('providers', self.providerID) if x['url'] == url]):
            return

        # ignore invalid urls
        if not validate_url(url) and not url.startswith('magnet') \
                or is_ip_private(url.split(r'//')[-1].split(r'/')[0]):
            return

        try:
            # parse release name
            parse_result = NameParser(validate_show=sickrage.app.config.enable_rss_cache_valid_shows).parse(name)
            if parse_result.series_name and parse_result.quality != Quality.UNKNOWN:
                season = parse_result.season_number if parse_result.season_number else 1
                episodes = parse_result.episode_numbers

                if season and episodes:
                    # store episodes as a seperated string
                    episodeText = "|" + "|".join(map(str, episodes)) + "|"

                    # get quality of release
                    quality = parse_result.quality

                    # get release group
                    release_group = parse_result.release_group

                    # get version
                    version = parse_result.version

                    dbData = {
                        '_t': 'providers',
                        'provider': self.providerID,
                        'name': name,
                        'season': season,
                        'episodes': episodeText,
                        'indexerid': parse_result.indexerid,
                        'url': url,
                        'time': int(time.mktime(datetime.datetime.today().timetuple())),
                        'quality': quality,
                        'release_group': release_group,
                        'version': version,
                        'seeders': seeders,
                        'leechers': leechers,
                        'size': size
                    }

                    # add to internal database
                    sickrage.app.cache_db.insert(dbData)

                    # add to external provider cache database
                    if sickrage.app.config.enable_api_providers_cache and not self.provider.private:
                        try:
                            sickrage.app.api.add_provider_cache_result(dbData)
                        except Exception:
                            pass

                    sickrage.app.log.debug("SEARCH RESULT:[%s] ADDED TO CACHE!", name)
        except (InvalidShowException, InvalidNameException):
            pass

    def search_cache(self, ep_obj, manualSearch=False, downCurQuality=False):
        neededEps = {}
        dbData = []

        # get data from external database
        if sickrage.app.config.enable_api_providers_cache and not self.provider.private:
            try:
                dbData += sickrage.app.api.get_provider_cache_results(self.providerID, ep_obj.show.indexerid)
            except Exception:
                pass

        # get data from internal database
        dbData += [x for x in sickrage.app.cache_db.get_many('providers', self.providerID)]

        # for each cache entry
        for curResult in (x for x in dbData if x['indexerid'] == ep_obj.show.indexerid and x['season'] == ep_obj.season
                                               and "|" + str(ep_obj.episode) + "|" in x['episodes']):
            result = self.provider.getResult()

            # ignore invalid urls
            if not validate_url(curResult["url"]) and not curResult["url"].startswith('magnet') \
                    or is_ip_private(curResult["url"].split(r'//')[-1].split(r'/')[0]):
                continue

            # ignored/required words, and non-tv junk
            if not show_names.filterBadReleases(curResult["name"]):
                continue

            # get the show object, or if it's not one of our shows then ignore it
            result.show = findCertainShow(int(curResult["indexerid"]))
            if not result.show:
                continue

            # skip if provider is anime only and show is not anime
            if self.provider.anime_only and not result.show.is_anime:
                sickrage.app.log.debug("" + str(result.show.name) + " is not an anime, skiping")
                continue

            # get season and ep data (ignoring multi-eps for now)
            curSeason = int(curResult["season"])
            if curSeason == -1:
                continue

            try:
                result.episodes = [result.show.getEpisode(curSeason, int(curEp)) for curEp in
                                   filter(None, curResult["episodes"].split("|"))]
            except EpisodeNotFoundException:
                continue

            result.quality = int(curResult["quality"])
            result.release_group = curResult["release_group"]
            result.version = curResult["version"]

            # make sure we want the episode
            wantEp = False
            for curEp in result.episodes:
                if result.show.wantEpisode(curEp.season, curEp.episode, result.quality, manualSearch, downCurQuality):
                    wantEp = True

            if not wantEp:
                sickrage.app.log.info("Skipping " + curResult["name"] + " because we don't want an episode that's " +
                                      Quality.qualityStrings[result.quality])
                continue

            # build a result object
            result.name = curResult["name"]
            result.url = curResult["url"]

            sickrage.app.log.info("Found result " + result.name + " at " + result.url)

            result.seeders = curResult.get("seeders", -1)
            result.leechers = curResult.get("leechers", -1)
            result.size = curResult.get("size", -1)
            result.content = None

            # add it to the list
            if ep_obj.episode not in neededEps:
                neededEps[ep_obj.episode] = [result]
            else:
                neededEps[ep_obj.episode] += [result]

        # datetime stamp this search so cache gets cleared
        self.last_search = datetime.datetime.today()

        return neededEps



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals
from __future__ import unicode_literals

import re
import string
import struct

__all__ = ['resolve']


def resolve(code):
    """
    Transform a twocc or fourcc code into a name.  Returns a 2-tuple of (cc,
    codec) where both are strings and cc is a string in the form '0xXX' if it's
    a twocc, or 'ABCD' if it's a fourcc.  If the given code is not a known
    twocc or fourcc, the return value will be (None, 'Unknown'), unless the
    code is otherwise a printable string in which case it will be returned as
    the codec.
    """
    if isinstance(code, basestring):
        codec = 'Unknown'
        # Check for twocc
        if re.match(r'^0x[\da-f]{1,4}$', code, re.I):
            # Twocc in hex form
            return code, TWOCC.get(int(code, 16), codec)
        elif code.isdigit() and 0 <= int(code) <= 0xff:
            # Twocc in decimal form
            return hex(int(code)), TWOCC.get(int(code), codec)
        elif len(code) == 2:
            code = struct.unpack('H', code)[0]
            return hex(code), TWOCC.get(code, codec)
        elif len(code) != 4 and len([x for x in code if x not in string.printable]) == 0:
            # Code is a printable string.
            codec = unicode(code)

        if code[:2] == 'MS' and code[2:].upper() in FOURCC:
            code = code[2:]

        if code.upper() in FOURCC:
            return code.upper(), unicode(FOURCC[code.upper()])
        return None, codec
    elif isinstance(code, (int, long)):
        return hex(code), TWOCC.get(code, 'Unknown')

    return None, 'Unknown'


TWOCC = {
    0x0000: 'Unknown Wave Format',
    0x0001: 'PCM',
    0x0002: 'Microsoft ADPCM',
    0x0003: 'IEEE Float',
    0x0004: 'Compaq Computer VSELP',
    0x0005: 'IBM CVSD',
    0x0006: 'A-Law',
    0x0007: 'mu-Law',
    0x0008: 'Microsoft DTS',
    0x0009: 'Microsoft DRM',
    0x0010: 'OKI ADPCM',
    0x0011: 'Intel DVI/IMA ADPCM',
    0x0012: 'Videologic MediaSpace ADPCM',
    0x0013: 'Sierra Semiconductor ADPCM',
    0x0014: 'Antex Electronics G.723 ADPCM',
    0x0015: 'DSP Solutions DigiSTD',
    0x0016: 'DSP Solutions DigiFIX',
    0x0017: 'Dialogic OKI ADPCM',
    0x0018: 'MediaVision ADPCM',
    0x0019: 'Hewlett-Packard CU',
    0x0020: 'Yamaha ADPCM',
    0x0021: 'Speech Compression Sonarc',
    0x0022: 'DSP Group TrueSpeech',
    0x0023: 'Echo Speech EchoSC1',
    0x0024: 'Audiofile AF36',
    0x0025: 'Audio Processing Technology APTX',
    0x0026: 'AudioFile AF10',
    0x0027: 'Prosody 1612',
    0x0028: 'LRC',
    0x0030: 'Dolby AC2',
    0x0031: 'Microsoft GSM 6.10',
    0x0032: 'MSNAudio',
    0x0033: 'Antex Electronics ADPCME',
    0x0034: 'Control Resources VQLPC',
    0x0035: 'DSP Solutions DigiREAL',
    0x0036: 'DSP Solutions DigiADPCM',
    0x0037: 'Control Resources CR10',
    0x0038: 'Natural MicroSystems VBXADPCM',
    0x0039: 'Crystal Semiconductor IMA ADPCM',
    0x003A: 'EchoSC3',
    0x003B: 'Rockwell ADPCM',
    0x003C: 'Rockwell Digit LK',
    0x003D: 'Xebec',
    0x0040: 'Antex Electronics G.721 ADPCM',
    0x0041: 'G.728 CELP',
    0x0042: 'MSG723',
    0x0043: 'IBM AVC ADPCM',
    0x0045: 'ITU-T G.726 ADPCM',
    0x0050: 'MPEG 1, Layer 1,2',
    0x0052: 'RT24',
    0x0053: 'PAC',
    0x0055: 'MPEG Layer 3',
    0x0059: 'Lucent G.723',
    0x0060: 'Cirrus',
    0x0061: 'ESPCM',
    0x0062: 'Voxware',
    0x0063: 'Canopus Atrac',
    0x0064: 'G.726 ADPCM',
    0x0065: 'G.722 ADPCM',
    0x0066: 'DSAT',
    0x0067: 'DSAT Display',
    0x0069: 'Voxware Byte Aligned',
    0x0070: 'Voxware AC8',
    0x0071: 'Voxware AC10',
    0x0072: 'Voxware AC16',
    0x0073: 'Voxware AC20',
    0x0074: 'Voxware MetaVoice',
    0x0075: 'Voxware MetaSound',
    0x0076: 'Voxware RT29HW',
    0x0077: 'Voxware VR12',
    0x0078: 'Voxware VR18',
    0x0079: 'Voxware TQ40',
    0x0080: 'Softsound',
    0x0081: 'Voxware TQ60',
    0x0082: 'MSRT24',
    0x0083: 'G.729A',
    0x0084: 'MVI MV12',
    0x0085: 'DF G.726',
    0x0086: 'DF GSM610',
    0x0088: 'ISIAudio',
    0x0089: 'Onlive',
    0x0091: 'SBC24',
    0x0092: 'Dolby AC3 SPDIF',
    0x0093: 'MediaSonic G.723',
    0x0094: 'Aculab PLC Prosody 8KBPS',
    0x0097: 'ZyXEL ADPCM',
    0x0098: 'Philips LPCBB',
    0x0099: 'Packed',
    0x00A0: 'Malden Electronics PHONYTALK',
    0x00FF: 'AAC',
    0x0100: 'Rhetorex ADPCM',
    0x0101: 'IBM mu-law',
    0x0102: 'IBM A-law',
    0x0103: 'IBM AVC Adaptive Differential Pulse Code Modulation',
    0x0111: 'Vivo G.723',
    0x0112: 'Vivo Siren',
    0x0123: 'Digital G.723',
    0x0125: 'Sanyo LD ADPCM',
    0x0130: 'Sipro Lab Telecom ACELP.net',
    0x0131: 'Sipro Lab Telecom ACELP.4800',
    0x0132: 'Sipro Lab Telecom ACELP.8V3',
    0x0133: 'Sipro Lab Telecom ACELP.G.729',
    0x0134: 'Sipro Lab Telecom ACELP.G.729A',
    0x0135: 'Sipro Lab Telecom ACELP.KELVIN',
    0x0140: 'Windows Media Video V8',
    0x0150: 'Qualcomm PureVoice',
    0x0151: 'Qualcomm HalfRate',
    0x0155: 'Ring Zero Systems TUB GSM',
    0x0160: 'Windows Media Audio V1 / DivX audio (WMA)',
    0x0161: 'Windows Media Audio V7 / V8 / V9',
    0x0162: 'Windows Media Audio Professional V9',
    0x0163: 'Windows Media Audio Lossless V9',
    0x0170: 'UNISYS NAP ADPCM',
    0x0171: 'UNISYS NAP ULAW',
    0x0172: 'UNISYS NAP ALAW',
    0x0173: 'UNISYS NAP 16K',
    0x0200: 'Creative Labs ADPCM',
    0x0202: 'Creative Labs Fastspeech8',
    0x0203: 'Creative Labs Fastspeech10',
    0x0210: 'UHER Informatic ADPCM',
    0x0215: 'Ulead DV ACM',
    0x0216: 'Ulead DV ACM',
    0x0220: 'Quarterdeck',
    0x0230: 'I-link Worldwide ILINK VC',
    0x0240: 'Aureal Semiconductor RAW SPORT',
    0x0241: 'ESST AC3',
    0x0250: 'Interactive Products HSX',
    0x0251: 'Interactive Products RPELP',
    0x0260: 'Consistent Software CS2',
    0x0270: 'Sony ATRAC3 (SCX, same as MiniDisk LP2)',
    0x0300: 'Fujitsu FM Towns Snd',
    0x0400: 'BTV Digital',
    0x0401: 'Intel Music Coder (IMC)',
    0x0402: 'Ligos Indeo Audio',
    0x0450: 'QDesign Music',
    0x0680: 'VME VMPCM',
    0x0681: 'AT&T Labs TPC',
    0x0700: 'YMPEG Alpha',
    0x08AE: 'ClearJump LiteWave',
    0x1000: 'Olivetti GSM',
    0x1001: 'Olivetti ADPCM',
    0x1002: 'Olivetti CELP',
    0x1003: 'Olivetti SBC',
    0x1004: 'Olivetti OPR',
    0x1100: 'Lernout & Hauspie LH Codec',
    0x1101: 'Lernout & Hauspie CELP codec',
    0x1102: 'Lernout & Hauspie SBC codec',
    0x1103: 'Lernout & Hauspie SBC codec',
    0x1104: 'Lernout & Hauspie SBC codec',
    0x1400: 'Norris',
    0x1401: 'AT&T ISIAudio',
    0x1500: 'Soundspace Music Compression',
    0x181C: 'VoxWare RT24 speech codec',
    0x181E: 'Lucent elemedia AX24000P Music codec',
    0x1C07: 'Lucent SX8300P speech codec',
    0x1C0C: 'Lucent SX5363S G.723 compliant codec',
    0x1F03: 'CUseeMe DigiTalk (ex-Rocwell)',
    0x1FC4: 'NCT Soft ALF2CD ACM',
    0x2000: 'AC3',
    0x2001: 'Dolby DTS (Digital Theater System)',
    0x2002: 'RealAudio 1 / 2 14.4',
    0x2003: 'RealAudio 1 / 2 28.8',
    0x2004: 'RealAudio G2 / 8 Cook (low bitrate)',
    0x2005: 'RealAudio 3 / 4 / 5 Music (DNET)',
    0x2006: 'RealAudio 10 AAC (RAAC)',
    0x2007: 'RealAudio 10 AAC+ (RACP)',
    0x3313: 'makeAVIS',
    0x4143: 'Divio MPEG-4 AAC audio',
    0x434C: 'LEAD Speech',
    0x564C: 'LEAD Vorbis',
    0x674F: 'Ogg Vorbis (mode 1)',
    0x6750: 'Ogg Vorbis (mode 2)',
    0x6751: 'Ogg Vorbis (mode 3)',
    0x676F: 'Ogg Vorbis (mode 1+)',
    0x6770: 'Ogg Vorbis (mode 2+)',
    0x6771: 'Ogg Vorbis (mode 3+)',
    0x7A21: 'GSM-AMR (CBR, no SID)',
    0x7A22: 'GSM-AMR (VBR, including SID)',
    0xDFAC: 'DebugMode SonicFoundry Vegas FrameServer ACM Codec',
    0xF1AC: 'Free Lossless Audio Codec FLAC',
    0xFFFE: 'Extensible wave format',
    0xFFFF: 'development'
}


FOURCC = {
    '1978': 'A.M.Paredes predictor (LossLess)',
    '2VUY': 'Optibase VideoPump 8-bit 4:2:2 Component YCbCr',
    '3IV0': 'MPEG4-based codec 3ivx',
    '3IV1': '3ivx v1',
    '3IV2': '3ivx v2',
    '3IVD': 'FFmpeg DivX ;-) (MS MPEG-4 v3)',
    '3IVX': 'MPEG4-based codec 3ivx',
    '8BPS': 'Apple QuickTime Planar RGB with Alpha-channel',
    'AAS4': 'Autodesk Animator codec (RLE)',
    'AASC': 'Autodesk Animator',
    'ABYR': 'Kensington ABYR',
    'ACTL': 'Streambox ACT-L2',
    'ADV1': 'Loronix WaveCodec',
    'ADVJ': 'Avid M-JPEG Avid Technology Also known as AVRn',
    'AEIK': 'Intel Indeo Video 3.2',
    'AEMI': 'Array VideoONE MPEG1-I Capture',
    'AFLC': 'Autodesk Animator FLC',
    'AFLI': 'Autodesk Animator FLI',
    'AHDV': 'CineForm 10-bit Visually Perfect HD',
    'AJPG': '22fps JPEG-based codec for digital cameras',
    'AMPG': 'Array VideoONE MPEG',
    'ANIM': 'Intel RDX (ANIM)',
    'AP41': 'AngelPotion Definitive',
    'AP42': 'AngelPotion Definitive',
    'ASLC': 'AlparySoft Lossless Codec',
    'ASV1': 'Asus Video v1',
    'ASV2': 'Asus Video v2',
    'ASVX': 'Asus Video 2.0 (audio)',
    'ATM4': 'Ahead Nero Digital MPEG-4 Codec',
    'AUR2': 'Aura 2 Codec - YUV 4:2:2',
    'AURA': 'Aura 1 Codec - YUV 4:1:1',
    'AV1X': 'Avid 1:1x (Quick Time)',
    'AVC1': 'H.264 AVC',
    'AVD1': 'Avid DV (Quick Time)',
    'AVDJ': 'Avid Meridien JFIF with Alpha-channel',
    'AVDN': 'Avid DNxHD (Quick Time)',
    'AVDV': 'Avid DV',
    'AVI1': 'MainConcept Motion JPEG Codec',
    'AVI2': 'MainConcept Motion JPEG Codec',
    'AVID': 'Avid Motion JPEG',
    'AVIS': 'Wrapper for AviSynth',
    'AVMP': 'Avid IMX (Quick Time)',
    'AVR ': 'Avid ABVB/NuVista MJPEG with Alpha-channel',
    'AVRN': 'Avid Motion JPEG',
    'AVUI': 'Avid Meridien Uncompressed with Alpha-channel',
    'AVUP': 'Avid 10bit Packed (Quick Time)',
    'AYUV': '4:4:4 YUV (AYUV)',
    'AZPR': 'Quicktime Apple Video',
    'AZRP': 'Quicktime Apple Video',
    'BGR ': 'Uncompressed BGR32 8:8:8:8',
    'BGR(15)': 'Uncompressed BGR15 5:5:5',
    'BGR(16)': 'Uncompressed BGR16 5:6:5',
    'BGR(24)': 'Uncompressed BGR24 8:8:8',
    'BHIV': 'BeHere iVideo',
    'BINK': 'RAD Game Tools Bink Video',
    'BIT ': 'BI_BITFIELDS (Raw RGB)',
    'BITM': 'Microsoft H.261',
    'BLOX': 'Jan Jezabek BLOX MPEG Codec',
    'BLZ0': 'DivX for Blizzard Decoder Filter',
    'BT20': 'Conexant Prosumer Video',
    'BTCV': 'Conexant Composite Video Codec',
    'BTVC': 'Conexant Composite Video',
    'BW00': 'BergWave (Wavelet)',
    'BW10': 'Data Translation Broadway MPEG Capture',
    'BXBG': 'BOXX BGR',
    'BXRG': 'BOXX RGB',
    'BXY2': 'BOXX 10-bit YUV',
    'BXYV': 'BOXX YUV',
    'CC12': 'Intel YUV12',
    'CDV5': 'Canopus SD50/DVHD',
    'CDVC': 'Canopus DV',
    'CDVH': 'Canopus SD50/DVHD',
    'CFCC': 'Digital Processing Systems DPS Perception',
    'CFHD': 'CineForm 10-bit Visually Perfect HD',
    'CGDI': 'Microsoft Office 97 Camcorder Video',
    'CHAM': 'Winnov Caviara Champagne',
    'CJPG': 'Creative WebCam JPEG',
    'CLJR': 'Cirrus Logic YUV 4 pixels',
    'CLLC': 'Canopus LossLess',
    'CLPL': 'YV12',
    'CMYK': 'Common Data Format in Printing',
    'COL0': 'FFmpeg DivX ;-) (MS MPEG-4 v3)',
    'COL1': 'FFmpeg DivX ;-) (MS MPEG-4 v3)',
    'CPLA': 'Weitek 4:2:0 YUV Planar',
    'CRAM': 'Microsoft Video 1 (CRAM)',
    'CSCD': 'RenderSoft CamStudio lossless Codec',
    'CTRX': 'Citrix Scalable Video Codec',
    'CUVC': 'Canopus HQ',
    'CVID': 'Radius Cinepak',
    'CWLT': 'Microsoft Color WLT DIB',
    'CYUV': 'Creative Labs YUV',
    'CYUY': 'ATI YUV',
    'D261': 'H.261',
    'D263': 'H.263',
    'DAVC': 'Dicas MPEGable H.264/MPEG-4 AVC base profile codec',
    'DC25': 'MainConcept ProDV Codec',
    'DCAP': 'Pinnacle DV25 Codec',
    'DCL1': 'Data Connection Conferencing Codec',
    'DCT0': 'WniWni Codec',
    'DFSC': 'DebugMode FrameServer VFW Codec',
    'DIB ': 'Full Frames (Uncompressed)',
    'DIV1': 'FFmpeg-4 V1 (hacked MS MPEG-4 V1)',
    'DIV2': 'MS MPEG-4 V2',
    'DIV3': 'DivX v3 MPEG-4 Low-Motion',
    'DIV4': 'DivX v3 MPEG-4 Fast-Motion',
    'DIV5': 'DIV5',
    'DIV6': 'DivX MPEG-4',
    'DIVX': 'DivX',
    'DM4V': 'Dicas MPEGable MPEG-4',
    'DMB1': 'Matrox Rainbow Runner hardware MJPEG',
    'DMB2': 'Paradigm MJPEG',
    'DMK2': 'ViewSonic V36 PDA Video',
    'DP02': 'DynaPel MPEG-4',
    'DPS0': 'DPS Reality Motion JPEG',
    'DPSC': 'DPS PAR Motion JPEG',
    'DRWX': 'Pinnacle DV25 Codec',
    'DSVD': 'DSVD',
    'DTMT': 'Media-100 Codec',
    'DTNT': 'Media-100 Codec',
    'DUCK': 'Duck True Motion 1.0',
    'DV10': 'BlueFish444 (lossless RGBA, YUV 10-bit)',
    'DV25': 'Matrox DVCPRO codec',
    'DV50': 'Matrox DVCPRO50 codec',
    'DVAN': 'DVAN',
    'DVC ': 'Apple QuickTime DV (DVCPRO NTSC)',
    'DVCP': 'Apple QuickTime DV (DVCPRO PAL)',
    'DVCS': 'MainConcept DV Codec',
    'DVE2': 'InSoft DVE-2 Videoconferencing',
    'DVH1': 'Pinnacle DVHD100',
    'DVHD': 'DV 1125 lines at 30.00 Hz or 1250 lines at 25.00 Hz',
    'DVIS': 'VSYNC DualMoon Iris DV codec',
    'DVL ': 'Radius SoftDV 16:9 NTSC',
    'DVLP': 'Radius SoftDV 16:9 PAL',
    'DVMA': 'Darim Vision DVMPEG',
    'DVOR': 'BlueFish444 (lossless RGBA, YUV 10-bit)',
    'DVPN': 'Apple QuickTime DV (DV NTSC)',
    'DVPP': 'Apple QuickTime DV (DV PAL)',
    'DVR1': 'TARGA2000 Codec',
    'DVRS': 'VSYNC DualMoon Iris DV codec',
    'DVSD': 'DV',
    'DVSL': 'DV compressed in SD (SDL)',
    'DVX1': 'DVX1000SP Video Decoder',
    'DVX2': 'DVX2000S Video Decoder',
    'DVX3': 'DVX3000S Video Decoder',
    'DX50': 'DivX v5',
    'DXGM': 'Electronic Arts Game Video codec',
    'DXSB': 'DivX Subtitles Codec',
    'DXT1': 'Microsoft DirectX Compressed Texture (DXT1)',
    'DXT2': 'Microsoft DirectX Compressed Texture (DXT2)',
    'DXT3': 'Microsoft DirectX Compressed Texture (DXT3)',
    'DXT4': 'Microsoft DirectX Compressed Texture (DXT4)',
    'DXT5': 'Microsoft DirectX Compressed Texture (DXT5)',
    'DXTC': 'Microsoft DirectX Compressed Texture (DXTC)',
    'DXTN': 'Microsoft DirectX Compressed Texture (DXTn)',
    'EKQ0': 'Elsa EKQ0',
    'ELK0': 'Elsa ELK0',
    'EM2V': 'Etymonix MPEG-2 I-frame',
    'EQK0': 'Elsa graphics card quick codec',
    'ESCP': 'Eidos Escape',
    'ETV1': 'eTreppid Video ETV1',
    'ETV2': 'eTreppid Video ETV2',
    'ETVC': 'eTreppid Video ETVC',
    'FFDS': 'FFDShow supported',
    'FFV1': 'FFDShow supported',
    'FFVH': 'FFVH codec',
    'FLIC': 'Autodesk FLI/FLC Animation',
    'FLJP': 'D-Vision Field Encoded Motion JPEG',
    'FLV1': 'FLV1 codec',
    'FMJP': 'D-Vision fieldbased ISO MJPEG',
    'FRLE': 'SoftLab-NSK Y16 + Alpha RLE',
    'FRWA': 'SoftLab-Nsk Forward Motion JPEG w/ alpha channel',
    'FRWD': 'SoftLab-Nsk Forward Motion JPEG',
    'FRWT': 'SoftLab-NSK Vision Forward Motion JPEG with Alpha-channel',
    'FRWU': 'SoftLab-NSK Vision Forward Uncompressed',
    'FVF1': 'Iterated Systems Fractal Video Frame',
    'FVFW': 'ff MPEG-4 based on XviD codec',
    'GEPJ': 'White Pine (ex Paradigm Matrix) Motion JPEG Codec',
    'GJPG': 'Grand Tech GT891x Codec',
    'GLCC': 'GigaLink AV Capture codec',
    'GLZW': 'Motion LZW',
    'GPEG': 'Motion JPEG',
    'GPJM': 'Pinnacle ReelTime MJPEG Codec',
    'GREY': 'Apparently a duplicate of Y800',
    'GWLT': 'Microsoft Greyscale WLT DIB',
    'H260': 'H.260',
    'H261': 'H.261',
    'H262': 'H.262',
    'H263': 'H.263',
    'H264': 'H.264 AVC',
    'H265': 'H.265 HEVC',
    'H266': 'H.266',
    'H267': 'H.267',
    'H268': 'H.268',
    'H269': 'H.269',
    'HD10': 'BlueFish444 (lossless RGBA, YUV 10-bit)',
    'HDX4': 'Jomigo HDX4',
    'HEVC': 'H.265 HEVC',
    'HFYU': 'Huffman Lossless Codec',
    'HMCR': 'Rendition Motion Compensation Format (HMCR)',
    'HMRR': 'Rendition Motion Compensation Format (HMRR)',
    'I263': 'Intel ITU H.263 Videoconferencing (i263)',
    'I420': 'Intel Indeo 4',
    'IAN ': 'Intel RDX',
    'ICLB': 'InSoft CellB Videoconferencing',
    'IDM0': 'IDM Motion Wavelets 2.0',
    'IF09': 'Microsoft H.261',
    'IGOR': 'Power DVD',
    'IJPG': 'Intergraph JPEG',
    'ILVC': 'Intel Layered Video',
    'ILVR': 'ITU-T H.263+',
    'IMC1': 'IMC1',
    'IMC2': 'IMC2',
    'IMC3': 'IMC3',
    'IMC4': 'IMC4',
    'IMJG': 'Accom SphereOUS MJPEG with Alpha-channel',
    'IPDV': 'I-O Data Device Giga AVI DV Codec',
    'IPJ2': 'Image Power JPEG2000',
    'IR21': 'Intel Indeo 2.1',
    'IRAW': 'Intel YUV Uncompressed',
    'IUYV': 'Interlaced version of UYVY (line order 0,2,4 then 1,3,5 etc)',
    'IV30': 'Ligos Indeo 3.0',
    'IV31': 'Ligos Indeo 3.1',
    'IV32': 'Ligos Indeo 3.2',
    'IV33': 'Ligos Indeo 3.3',
    'IV34': 'Ligos Indeo 3.4',
    'IV35': 'Ligos Indeo 3.5',
    'IV36': 'Ligos Indeo 3.6',
    'IV37': 'Ligos Indeo 3.7',
    'IV38': 'Ligos Indeo 3.8',
    'IV39': 'Ligos Indeo 3.9',
    'IV40': 'Ligos Indeo Interactive 4.0',
    'IV41': 'Ligos Indeo Interactive 4.1',
    'IV42': 'Ligos Indeo Interactive 4.2',
    'IV43': 'Ligos Indeo Interactive 4.3',
    'IV44': 'Ligos Indeo Interactive 4.4',
    'IV45': 'Ligos Indeo Interactive 4.5',
    'IV46': 'Ligos Indeo Interactive 4.6',
    'IV47': 'Ligos Indeo Interactive 4.7',
    'IV48': 'Ligos Indeo Interactive 4.8',
    'IV49': 'Ligos Indeo Interactive 4.9',
    'IV50': 'Ligos Indeo Interactive 5.0',
    'IY41': 'Interlaced version of Y41P (line order 0,2,4,...,1,3,5...)',
    'IYU1': '12 bit format used in mode 2 of the IEEE 1394 Digital Camera 1.04 spec',
    'IYU2': '24 bit format used in mode 2 of the IEEE 1394 Digital Camera 1.04 spec',
    'IYUV': 'Intel Indeo iYUV 4:2:0',
    'JBYR': 'Kensington JBYR',
    'JFIF': 'Motion JPEG (FFmpeg)',
    'JPEG': 'Still Image JPEG DIB',
    'JPG ': 'JPEG compressed',
    'JPGL': 'Webcam JPEG Light',
    'KMVC': 'Karl Morton\'s Video Codec',
    'KPCD': 'Kodak Photo CD',
    'L261': 'Lead Technologies H.261',
    'L263': 'Lead Technologies H.263',
    'LAGS': 'Lagarith LossLess',
    'LBYR': 'Creative WebCam codec',
    'LCMW': 'Lead Technologies Motion CMW Codec',
    'LCW2': 'LEADTools MCMW 9Motion Wavelet)',
    'LEAD': 'LEAD Video Codec',
    'LGRY': 'Lead Technologies Grayscale Image',
    'LJ2K': 'LEADTools JPEG2000',
    'LJPG': 'LEAD MJPEG Codec',
    'LMP2': 'LEADTools MPEG2',
    'LOCO': 'LOCO Lossless Codec',
    'LSCR': 'LEAD Screen Capture',
    'LSVM': 'Vianet Lighting Strike Vmail (Streaming)',
    'LZO1': 'LZO compressed (lossless codec)',
    'M261': 'Microsoft H.261',
    'M263': 'Microsoft H.263',
    'M4CC': 'ESS MPEG4 Divio codec',
    'M4S2': 'Microsoft MPEG-4 (M4S2)',
    'MC12': 'ATI Motion Compensation Format (MC12)',
    'MC24': 'MainConcept Motion JPEG Codec',
    'MCAM': 'ATI Motion Compensation Format (MCAM)',
    'MCZM': 'Theory MicroCosm Lossless 64bit RGB with Alpha-channel',
    'MDVD': 'Alex MicroDVD Video (hacked MS MPEG-4)',
    'MDVF': 'Pinnacle DV/DV50/DVHD100',
    'MHFY': 'A.M.Paredes mhuffyYUV (LossLess)',
    'MJ2C': 'Morgan Multimedia Motion JPEG2000',
    'MJPA': 'Pinnacle ReelTime MJPG hardware codec',
    'MJPB': 'Motion JPEG codec',
    'MJPG': 'Motion JPEG DIB',
    'MJPX': 'Pegasus PICVideo Motion JPEG',
    'MMES': 'Matrox MPEG-2 I-frame',
    'MNVD': 'MindBend MindVid LossLess',
    'MP2A': 'MPEG-2 Audio',
    'MP2T': 'MPEG-2 Transport Stream',
    'MP2V': 'MPEG-2 Video',
    'MP41': 'Microsoft MPEG-4 V1 (enhansed H263)',
    'MP42': 'Microsoft MPEG-4 (low-motion)',
    'MP43': 'Microsoft MPEG-4 (fast-motion)',
    'MP4A': 'MPEG-4 Audio',
    'MP4S': 'Microsoft MPEG-4 (MP4S)',
    'MP4T': 'MPEG-4 Transport Stream',
    'MP4V': 'Apple QuickTime MPEG-4 native',
    'MPEG': 'MPEG-1',
    'MPG1': 'FFmpeg-1',
    'MPG2': 'FFmpeg-1',
    'MPG3': 'Same as Low motion DivX MPEG-4',
    'MPG4': 'Microsoft MPEG-4 Video High Speed Compressor',
    'MPGI': 'Sigma Designs MPEG',
    'MPNG': 'Motion PNG codec',
    'MRCA': 'Martin Regen Codec',
    'MRLE': 'Run Length Encoding',
    'MSS1': 'Windows Screen Video',
    'MSS2': 'Windows Media 9',
    'MSUC': 'MSU LossLess',
    'MSVC': 'Microsoft Video 1',
    'MSZH': 'Lossless codec (ZIP compression)',
    'MTGA': 'Motion TGA images (24, 32 bpp)',
    'MTX1': 'Matrox MTX1',
    'MTX2': 'Matrox MTX2',
    'MTX3': 'Matrox MTX3',
    'MTX4': 'Matrox MTX4',
    'MTX5': 'Matrox MTX5',
    'MTX6': 'Matrox MTX6',
    'MTX7': 'Matrox MTX7',
    'MTX8': 'Matrox MTX8',
    'MTX9': 'Matrox MTX9',
    'MV12': 'MV12',
    'MVI1': 'Motion Pixels MVI',
    'MVI2': 'Motion Pixels MVI',
    'MWV1': 'Aware Motion Wavelets',
    'MYUV': 'Media-100 844/X Uncompressed',
    'NAVI': 'nAVI',
    'NDIG': 'Ahead Nero Digital MPEG-4 Codec',
    'NHVU': 'NVidia Texture Format (GEForce 3)',
    'NO16': 'Theory None16 64bit uncompressed RAW',
    'NT00': 'NewTek LigtWave HDTV YUV with Alpha-channel',
    'NTN1': 'Nogatech Video Compression 1',
    'NTN2': 'Nogatech Video Compression 2 (GrabBee hardware coder)',
    'NUV1': 'NuppelVideo',
    'NV12': '8-bit Y plane followed by an interleaved U/V plane with 2x2 subsampling',
    'NV21': 'As NV12 with U and V reversed in the interleaved plane',
    'NVDS': 'nVidia Texture Format',
    'NVHS': 'NVidia Texture Format (GEForce 3)',
    'NVS0': 'nVidia GeForce Texture',
    'NVS1': 'nVidia GeForce Texture',
    'NVS2': 'nVidia GeForce Texture',
    'NVS3': 'nVidia GeForce Texture',
    'NVS4': 'nVidia GeForce Texture',
    'NVS5': 'nVidia GeForce Texture',
    'NVT0': 'nVidia GeForce Texture',
    'NVT1': 'nVidia GeForce Texture',
    'NVT2': 'nVidia GeForce Texture',
    'NVT3': 'nVidia GeForce Texture',
    'NVT4': 'nVidia GeForce Texture',
    'NVT5': 'nVidia GeForce Texture',
    'PDVC': 'I-O Data Device Digital Video Capture DV codec',
    'PGVV': 'Radius Video Vision',
    'PHMO': 'IBM Photomotion',
    'PIM1': 'Pegasus Imaging',
    'PIM2': 'Pegasus Imaging',
    'PIMJ': 'Pegasus Imaging Lossless JPEG',
    'PIXL': 'MiroVideo XL (Motion JPEG)',
    'PNG ': 'Apple PNG',
    'PNG1': 'Corecodec.org CorePNG Codec',
    'PVEZ': 'Horizons Technology PowerEZ',
    'PVMM': 'PacketVideo Corporation MPEG-4',
    'PVW2': 'Pegasus Imaging Wavelet Compression',
    'PVWV': 'Pegasus Imaging Wavelet 2000',
    'PXLT': 'Apple Pixlet (Wavelet)',
    'Q1.0': 'Q-Team QPEG 1.0 (www.q-team.de)',
    'Q1.1': 'Q-Team QPEG 1.1 (www.q-team.de)',
    'QDGX': 'Apple QuickDraw GX',
    'QPEG': 'Q-Team QPEG 1.0',
    'QPEQ': 'Q-Team QPEG 1.1',
    'R210': 'BlackMagic YUV (Quick Time)',
    'R411': 'Radius DV NTSC YUV',
    'R420': 'Radius DV PAL YUV',
    'RAVI': 'GroupTRON ReferenceAVI codec (dummy for MPEG compressor)',
    'RAV_': 'GroupTRON ReferenceAVI codec (dummy for MPEG compressor)',
    'RAW ': 'Full Frames (Uncompressed)',
    'RGB ': 'Full Frames (Uncompressed)',
    'RGB(15)': 'Uncompressed RGB15 5:5:5',
    'RGB(16)': 'Uncompressed RGB16 5:6:5',
    'RGB(24)': 'Uncompressed RGB24 8:8:8',
    'RGB1': 'Uncompressed RGB332 3:3:2',
    'RGBA': 'Raw RGB with alpha',
    'RGBO': 'Uncompressed RGB555 5:5:5',
    'RGBP': 'Uncompressed RGB565 5:6:5',
    'RGBQ': 'Uncompressed RGB555X 5:5:5 BE',
    'RGBR': 'Uncompressed RGB565X 5:6:5 BE',
    'RGBT': 'Computer Concepts 32-bit support',
    'RL4 ': 'RLE 4bpp RGB',
    'RL8 ': 'RLE 8bpp RGB',
    'RLE ': 'Microsoft Run Length Encoder',
    'RLE4': 'Run Length Encoded 4',
    'RLE8': 'Run Length Encoded 8',
    'RMP4': 'REALmagic MPEG-4 Video Codec',
    'ROQV': 'Id RoQ File Video Decoder',
    'RPZA': 'Apple Video 16 bit "road pizza"',
    'RT21': 'Intel Real Time Video 2.1',
    'RTV0': 'NewTek VideoToaster',
    'RUD0': 'Rududu video codec',
    'RV10': 'RealVideo codec',
    'RV13': 'RealVideo codec',
    'RV20': 'RealVideo G2',
    'RV30': 'RealVideo 8',
    'RV40': 'RealVideo 9',
    'RVX ': 'Intel RDX (RVX )',
    'S263': 'Sorenson Vision H.263',
    'S422': 'Tekram VideoCap C210 YUV 4:2:2',
    'SAMR': 'Adaptive Multi-Rate (AMR) audio codec',
    'SAN3': 'MPEG-4 codec (direct copy of DivX 3.11a)',
    'SDCC': 'Sun Communication Digital Camera Codec',
    'SEDG': 'Samsung MPEG-4 codec',
    'SFMC': 'CrystalNet Surface Fitting Method',
    'SHR0': 'BitJazz SheerVideo',
    'SHR1': 'BitJazz SheerVideo',
    'SHR2': 'BitJazz SheerVideo',
    'SHR3': 'BitJazz SheerVideo',
    'SHR4': 'BitJazz SheerVideo',
    'SHR5': 'BitJazz SheerVideo',
    'SHR6': 'BitJazz SheerVideo',
    'SHR7': 'BitJazz SheerVideo',
    'SJPG': 'CUseeMe Networks Codec',
    'SL25': 'SoftLab-NSK DVCPRO',
    'SL50': 'SoftLab-NSK DVCPRO50',
    'SLDV': 'SoftLab-NSK Forward DV Draw codec',
    'SLIF': 'SoftLab-NSK MPEG2 I-frames',
    'SLMJ': 'SoftLab-NSK Forward MJPEG',
    'SMC ': 'Apple Graphics (SMC) codec (256 color)',
    'SMSC': 'Radius SMSC',
    'SMSD': 'Radius SMSD',
    'SMSV': 'WorldConnect Wavelet Video',
    'SNOW': 'SNOW codec',
    'SP40': 'SunPlus YUV',
    'SP44': 'SunPlus Aiptek MegaCam Codec',
    'SP53': 'SunPlus Aiptek MegaCam Codec',
    'SP54': 'SunPlus Aiptek MegaCam Codec',
    'SP55': 'SunPlus Aiptek MegaCam Codec',
    'SP56': 'SunPlus Aiptek MegaCam Codec',
    'SP57': 'SunPlus Aiptek MegaCam Codec',
    'SP58': 'SunPlus Aiptek MegaCam Codec',
    'SPIG': 'Radius Spigot',
    'SPLC': 'Splash Studios ACM Audio Codec',
    'SPRK': 'Sorenson Spark',
    'SQZ2': 'Microsoft VXTreme Video Codec V2',
    'STVA': 'ST CMOS Imager Data (Bayer)',
    'STVB': 'ST CMOS Imager Data (Nudged Bayer)',
    'STVC': 'ST CMOS Imager Data (Bunched)',
    'STVX': 'ST CMOS Imager Data (Extended CODEC Data Format)',
    'STVY': 'ST CMOS Imager Data (Extended CODEC Data Format with Correction Data)',
    'SV10': 'Sorenson Video R1',
    'SVQ1': 'Sorenson Video R3',
    'SVQ3': 'Sorenson Video 3 (Apple Quicktime 5)',
    'SWC1': 'MainConcept Motion JPEG Codec',
    'T420': 'Toshiba YUV 4:2:0',
    'TGA ': 'Apple TGA (with Alpha-channel)',
    'THEO': 'FFVFW Supported Codec',
    'TIFF': 'Apple TIFF (with Alpha-channel)',
    'TIM2': 'Pinnacle RAL DVI',
    'TLMS': 'TeraLogic Motion Intraframe Codec (TLMS)',
    'TLST': 'TeraLogic Motion Intraframe Codec (TLST)',
    'TM20': 'Duck TrueMotion 2.0',
    'TM2A': 'Duck TrueMotion Archiver 2.0',
    'TM2X': 'Duck TrueMotion 2X',
    'TMIC': 'TeraLogic Motion Intraframe Codec (TMIC)',
    'TMOT': 'Horizons Technology TrueMotion S',
    'TR20': 'Duck TrueMotion RealTime 2.0',
    'TRLE': 'Akula Alpha Pro Custom AVI (LossLess)',
    'TSCC': 'TechSmith Screen Capture Codec',
    'TV10': 'Tecomac Low-Bit Rate Codec',
    'TVJP': 'TrueVision Field Encoded Motion JPEG',
    'TVMJ': 'Truevision TARGA MJPEG Hardware Codec',
    'TY0N': 'Trident TY0N',
    'TY2C': 'Trident TY2C',
    'TY2N': 'Trident TY2N',
    'U263': 'UB Video StreamForce H.263',
    'U<Y ': 'Discreet UC YUV 4:2:2:4 10 bit',
    'U<YA': 'Discreet UC YUV 4:2:2:4 10 bit (with Alpha-channel)',
    'UCOD': 'eMajix.com ClearVideo',
    'ULTI': 'IBM Ultimotion',
    'UMP4': 'UB Video MPEG 4',
    'UYNV': 'UYVY',
    'UYVP': 'YCbCr 4:2:2',
    'UYVU': 'SoftLab-NSK Forward YUV codec',
    'UYVY': 'UYVY 4:2:2 byte ordering',
    'V210': 'Optibase VideoPump 10-bit 4:2:2 Component YCbCr',
    'V261': 'Lucent VX2000S',
    'V422': '24 bit YUV 4:2:2 Format',
    'V655': '16 bit YUV 4:2:2 Format',
    'VBLE': 'MarcFD VBLE Lossless Codec',
    'VCR1': 'ATI VCR 1.0',
    'VCR2': 'ATI VCR 2.0',
    'VCR3': 'ATI VCR 3.0',
    'VCR4': 'ATI VCR 4.0',
    'VCR5': 'ATI VCR 5.0',
    'VCR6': 'ATI VCR 6.0',
    'VCR7': 'ATI VCR 7.0',
    'VCR8': 'ATI VCR 8.0',
    'VCR9': 'ATI VCR 9.0',
    'VDCT': 'Video Maker Pro DIB',
    'VDOM': 'VDOnet VDOWave',
    'VDOW': 'VDOnet VDOLive (H.263)',
    'VDST': 'VirtualDub remote frameclient ICM driver',
    'VDTZ': 'Darim Vison VideoTizer YUV',
    'VGPX': 'VGPixel Codec',
    'VIDM': 'DivX 5.0 Pro Supported Codec',
    'VIDS': 'YUV 4:2:2 CCIR 601 for V422',
    'VIFP': 'VIFP',
    'VIV1': 'Vivo H.263',
    'VIV2': 'Vivo H.263',
    'VIVO': 'Vivo H.263 v2.00',
    'VIXL': 'Miro Video XL',
    'VLV1': 'Videologic VLCAP.DRV',
    'VP30': 'On2 VP3.0',
    'VP31': 'On2 VP3.1',
    'VP40': 'On2 TrueCast VP4',
    'VP50': 'On2 TrueCast VP5',
    'VP60': 'On2 TrueCast VP6',
    'VP61': 'On2 TrueCast VP6.1',
    'VP62': 'On2 TrueCast VP6.2',
    'VP70': 'On2 TrueMotion VP7',
    'VQC1': 'Vector-quantised codec 1',
    'VQC2': 'Vector-quantised codec 2',
    'VR21': 'BlackMagic YUV (Quick Time)',
    'VSSH': 'Vanguard VSS H.264',
    'VSSV': 'Vanguard Software Solutions Video Codec',
    'VSSW': 'Vanguard VSS H.264',
    'VTLP': 'Alaris VideoGramPixel Codec',
    'VX1K': 'VX1000S Video Codec',
    'VX2K': 'VX2000S Video Codec',
    'VXSP': 'VX1000SP Video Codec',
    'VYU9': 'ATI Technologies YUV',
    'VYUY': 'ATI Packed YUV Data',
    'WBVC': 'Winbond W9960',
    'WHAM': 'Microsoft Video 1 (WHAM)',
    'WINX': 'Winnov Software Compression',
    'WJPG': 'AverMedia Winbond JPEG',
    'WMV1': 'Windows Media Video V7',
    'WMV2': 'Windows Media Video V8',
    'WMV3': 'Windows Media Video V9',
    'WMVA': 'WMVA codec',
    'WMVP': 'Windows Media Video V9',
    'WNIX': 'WniWni Codec',
    'WNV1': 'Winnov Hardware Compression',
    'WNVA': 'Winnov hw compress',
    'WRLE': 'Apple QuickTime BMP Codec',
    'WRPR': 'VideoTools VideoServer Client Codec',
    'WV1F': 'WV1F codec',
    'WVLT': 'IllusionHope Wavelet 9/7',
    'WVP2': 'WVP2 codec',
    'X263': 'Xirlink H.263',
    'X264': 'XiWave GNU GPL x264 MPEG-4 Codec',
    'X265': 'H.265 HEVC',
    'XLV0': 'NetXL Video Decoder',
    'XMPG': 'Xing MPEG (I-Frame only)',
    'XVID': 'XviD MPEG-4',
    'XVIX': 'Based on XviD MPEG-4 codec',
    'XWV0': 'XiWave Video Codec',
    'XWV1': 'XiWave Video Codec',
    'XWV2': 'XiWave Video Codec',
    'XWV3': 'XiWave Video Codec (Xi-3 Video)',
    'XWV4': 'XiWave Video Codec',
    'XWV5': 'XiWave Video Codec',
    'XWV6': 'XiWave Video Codec',
    'XWV7': 'XiWave Video Codec',
    'XWV8': 'XiWave Video Codec',
    'XWV9': 'XiWave Video Codec',
    'XXAN': 'XXAN',
    'XYZP': 'Extended PAL format XYZ palette',
    'Y211': 'YUV 2:1:1 Packed',
    'Y216': 'Pinnacle TARGA CineWave YUV (Quick Time)',
    'Y411': 'YUV 4:1:1 Packed',
    'Y41B': 'YUV 4:1:1 Planar',
    'Y41P': 'PC1 4:1:1',
    'Y41T': 'PC1 4:1:1 with transparency',
    'Y422': 'Y422',
    'Y42B': 'YUV 4:2:2 Planar',
    'Y42T': 'PCI 4:2:2 with transparency',
    'Y444': 'IYU2',
    'Y8  ': 'Grayscale video',
    'Y800': 'Simple grayscale video',
    'YC12': 'Intel YUV12 Codec',
    'YMPG': 'YMPEG Alpha',
    'YU12': 'ATI YV12 4:2:0 Planar',
    'YU92': 'Intel - YUV',
    'YUNV': 'YUNV',
    'YUV2': 'Apple Component Video (YUV 4:2:2)',
    'YUV8': 'Winnov Caviar YUV8',
    'YUV9': 'Intel YUV9',
    'YUVP': 'YCbCr 4:2:2',
    'YUY2': 'Uncompressed YUV 4:2:2',
    'YUYV': 'Canopus YUV',
    'YV12': 'YVU12 Planar',
    'YV16': 'Elecard YUV 4:2:2 Planar',
    'YV92': 'Intel Smart Video Recorder YVU9',
    'YVU9': 'Intel YVU9 Planar',
    'YVYU': 'YVYU 4:2:2 byte ordering',
    'ZLIB': 'ZLIB',
    'ZPEG': 'Metheus Video Zipper',
    'ZYGO': 'ZyGo Video Codec'
}

# make it fool prove
for code, value in FOURCC.items():
    if not code.upper() in FOURCC:
        FOURCC[code.upper()] = value
    if code.endswith(' '):
        FOURCC[code.strip().upper()] = value



# Author: Daniel Heimans
# URL: http://code.google.com/p/sickrage
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import math
import re
import socket
import time

import jsonrpclib

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import sanitizeSceneName
from sickrage.core.scene_exceptions import get_scene_exceptions
from sickrage.providers import TorrentProvider


class BTNProvider(TorrentProvider):
    def __init__(self):
        super(BTNProvider, self).__init__("BTN", 'http://api.broadcasthe.net', True)

        self.supports_absolute_numbering = True
        self.api_key = None
        self.reject_m2ts = False

        self.cache = BTNCache(self, min_time=15)

    def _check_auth(self):
        if not self.api_key:
            sickrage.app.log.warning("Invalid api key. Check your settings")

        return True

    def _check_auth_from_data(self, parsedJSON):
        if parsedJSON is None:
            return self._check_auth()

        return True

    def search(self, search_params, age=0, ep_obj=None):

        self._check_auth()

        results = []
        params = {}
        apikey = self.api_key

        # age in seconds
        if age:
            params['age'] = "<=" + str(int(age))

        if search_params:
            params.update(search_params)
            sickrage.app.log.debug("Search string: %s" % search_params)

        parsedJSON = self._api_call(apikey, params)
        if not parsedJSON:
            sickrage.app.log.debug("No data returned from provider")
            return results

        if self._check_auth_from_data(parsedJSON):

            found_torrents = {}
            if 'torrents' in parsedJSON:
                found_torrents = parsedJSON['torrents']

            # We got something, we know the API sends max 1000 results at a time.
            # See if there are more than 1000 results for our query, if not we
            # keep requesting until we've got everything.
            # max 150 requests per hour so limit at that. Scan every 15 minutes. 60 / 15 = 4.
            max_pages = 150
            results_per_page = 1000

            if 'results' in parsedJSON and int(parsedJSON['results']) >= results_per_page:
                pages_needed = int(math.ceil(int(parsedJSON['results']) / results_per_page))
                if pages_needed > max_pages:
                    pages_needed = max_pages

                # +1 because range(1,4) = 1, 2, 3
                for page in xrange(1, pages_needed + 1):
                    parsedJSON = self._api_call(apikey, params, results_per_page, page * results_per_page)
                    # Note that this these are individual requests and might time out individually. This would result in 'gaps'
                    # in the results. There is no way to fix this though.
                    if 'torrents' in parsedJSON:
                        found_torrents.update(parsedJSON['torrents'])

            for torrentid, torrent_info in found_torrents.items():
                if self.reject_m2ts and re.match(r'(?i)m2?ts', torrent_info.get('Container', '')):
                    continue

                (title, url) = self._get_title_and_url(torrent_info)
                if title and url:
                    sickrage.app.log.debug("Found result: {}".format(title))
                    results.append(torrent_info)

        # FIXME SORT RESULTS
        return results

    def _api_call(self, apikey, params=None, results_per_page=300, offset=0):
        parsedJSON = {}

        try:
            api = jsonrpclib.Server('http://' + self.urls['base_url'])
            parsedJSON = api.getTorrents(apikey, params or {}, int(results_per_page), int(offset))
            time.sleep(5)
        except jsonrpclib.jsonrpc.ProtocolError as e:
            if e.message[1] == 'Call Limit Exceeded':
                sickrage.app.log.warning("You have exceeded the limit of 150 calls per hour.")
            elif e.message[1] == 'Invalid API Key':
                sickrage.app.log.warning("Incorrect authentication credentials.")
            else:
                sickrage.app.log.error(
                    "JSON-RPC protocol error while accessing provider. Error: {}".format(e.message[1]))
        except (socket.error, socket.timeout, ValueError) as e:
            sickrage.app.log.warning("Error while accessing provider. Error: {}".format(e))

        return parsedJSON

    @staticmethod
    def _get_title_and_url(parsedJSON, **kwargs):

        # The BTN API gives a lot of information in response,
        # however SickRage is built mostly around Scene or
        # release names, which is why we are using them here.

        if 'ReleaseName' in parsedJSON and parsedJSON['ReleaseName']:
            title = parsedJSON['ReleaseName']

        else:
            # If we don't have a release name we need to get creative
            title = ''
            if 'Series' in parsedJSON:
                title += parsedJSON['Series']
            if 'GroupName' in parsedJSON:
                title += '.' + parsedJSON['GroupName'] if title else parsedJSON['GroupName']
            if 'Resolution' in parsedJSON:
                title += '.' + parsedJSON['Resolution'] if title else parsedJSON['Resolution']
            if 'Source' in parsedJSON:
                title += '.' + parsedJSON['Source'] if title else parsedJSON['Source']
            if 'Codec' in parsedJSON:
                title += '.' + parsedJSON['Codec'] if title else parsedJSON['Codec']
            if title:
                title = title.replace(' ', '.')

        url = ''
        if 'DownloadURL' in parsedJSON:
            url = parsedJSON['DownloadURL']
            if url:
                # unescaped / is valid in JSON, but it can be escaped
                url = url.replace("\\/", "/")

        return title, url

    @staticmethod
    def _get_season_search_strings(ep_obj, **kwargs):
        search_params = []
        current_params = {'category': 'Season'}

        # Search for entire seasons: no need to do special things for air by date or sports shows
        if ep_obj.show.air_by_date or ep_obj.show.sports:
            # Search for the year of the air by date show
            current_params['name'] = str(ep_obj.airdate).split('-')[0]
        elif ep_obj.show.is_anime:
            current_params['name'] = "%d" % ep_obj.scene_absolute_number
        else:
            current_params['name'] = 'Season ' + str(ep_obj.scene_season)

        # search
        if ep_obj.show.indexer == 1:
            current_params['tvdb'] = ep_obj.show.indexerid
            search_params.append(current_params)
        else:
            name_exceptions = list(
                set(get_scene_exceptions(ep_obj.show.indexerid) + [ep_obj.show.name]))
            for name in name_exceptions:
                # Search by name if we don't have tvdb id
                current_params['series'] = sanitizeSceneName(name)
                search_params.append(current_params)

        return search_params

    @staticmethod
    def _get_episode_search_strings(ep_obj, add_string='', **kwargs):

        if not ep_obj:
            return [{}]

        to_return = []
        search_params = {'category': 'Episode'}

        # episode
        if ep_obj.show.air_by_date or ep_obj.show.sports:
            date_str = str(ep_obj.airdate)

            # BTN uses dots in dates, we just search for the date since that
            # combined with the series identifier should result in just one episode
            search_params['name'] = date_str.replace('-', '.')
        elif ep_obj.show.anime:
            search_params['name'] = "%i" % int(ep_obj.scene_absolute_number)
        else:
            # Do a general name search for the episode, formatted like SXXEYY
            search_params['name'] = "S%02dE%02d" % (ep_obj.scene_season, ep_obj.scene_episode)

        # search
        if ep_obj.show.indexer == 1:
            search_params['tvdb'] = ep_obj.show.indexerid
            to_return.append(search_params)
        else:
            # add new query string for every exception
            name_exceptions = list(
                set(get_scene_exceptions(ep_obj.show.indexerid) + [ep_obj.show.name]))
            for cur_exception in name_exceptions:
                search_params['series'] = sanitizeSceneName(cur_exception)
                to_return.append(search_params)

        return to_return

    def _doGeneralSearch(self, search_string):
        # 'search' looks as broad is it can find. Can contain episode overview and title for example,
        # use with caution!
        return self.search({'search': search_string})


class BTNCache(TVCache):
    def _get_rss_data(self):
        # Get the torrents uploaded since last check.
        seconds_since_last_update = math.ceil(time.time() - time.mktime(self.last_update.timetuple()))

        # default to 15 minutes
        seconds_minTime = self.min_time * 60
        if seconds_since_last_update < seconds_minTime:
            seconds_since_last_update = seconds_minTime

        # Set maximum to 24 hours (24 * 60 * 60 = 86400 seconds) of "RSS" data search, older things will need to be done through backlog
        if seconds_since_last_update > 86400:
            sickrage.app.log.debug(
                "The last known successful update was more than 24 hours ago, only trying to fetch the last 24 hours!")
            seconds_since_last_update = 86400

        return {'entries': self.provider.search(search_params=None, age=seconds_since_last_update)}



# Authors:
# Derek Battams <derek@battams.ca>
# Pedro Jose Pereira Vieito (@pvieito) <pvieito@gmail.com>
#
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import re
import smtplib
import traceback
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate

import sickrage
from sickrage.notifiers import Notifiers


class EmailNotifier(Notifiers):
    def __init__(self):
        super(EmailNotifier, self).__init__()
        self.name = 'email'
        self.last_err = None

    def test_notify(self, host, port, smtp_from, use_tls, user, pwd, to):
        msg = MIMEText('This is a test message from SiCKRAGE.  If you\'re reading this, the test succeeded.')
        msg['Subject'] = 'SiCKRAGE: Test Message'
        msg['From'] = smtp_from
        msg['To'] = to
        msg['Date'] = formatdate(localtime=True)
        return self._sendmail(host, port, smtp_from, use_tls, user, pwd, [to], msg, True)

    def _notify_snatch(self, ep_name, title="Snatched:"):
        """
        Send a notification that an episode was snatched

        ep_name: The name of the episode that was snatched
        title: The title of the notification (optional)
        """
        ep_name = ep_name

        if sickrage.app.config.email_notify_onsnatch:
            show = self._parseEp(ep_name)
            to = self._generate_recipients(show)
            if len(to) == 0:
                sickrage.app.log.warning('Skipping email notify because there are no configured recipients')
            else:
                try:
                    msg = MIMEMultipart('alternative')
                    msg.attach(MIMEText(
                        "<body style='font-family:Helvetica, Arial, sans-serif;'><h3>SiCKRAGE Notification - Snatched</h3>\n<p>Show: <b>" + re.search(
                            "(.+?) -.+", ep_name).group(1) + "</b></p>\n<p>Episode: <b>" + re.search(
                            ".+ - (.+?-.+) -.+", ep_name).group(
                            1) + "</b></p>\n\n<footer style='margin-top: 2.5em; padding: .7em 0; color: #777; border-top: #BBB solid 1px;'>Powered by SiCKRAGE.</footer></body>",
                        'html'))
                except:
                    try:
                        msg = MIMEText(ep_name)
                    except:
                        msg = MIMEText("Episode Snatched")

                msg['Subject'] = 'Snatched: ' + ep_name
                msg['From'] = sickrage.app.config.email_from
                msg['To'] = ','.join(to)
                msg['Date'] = formatdate(localtime=True)
                if self._sendmail(sickrage.app.config.email_host, sickrage.app.config.email_port,
                                  sickrage.app.config.email_from,
                                  sickrage.app.config.email_tls,
                                  sickrage.app.config.email_user, sickrage.app.config.email_password, to,
                                  msg):
                    sickrage.app.log.debug("Snatch notification sent to [%s] for '%s'" % (to, ep_name))
                else:
                    sickrage.app.log.error("Snatch notification ERROR: %s" % self.last_err)

    def _notify_download(self, ep_name, title="Completed:"):
        """
        Send a notification that an episode was downloaded

        ep_name: The name of the episode that was downloaded
        title: The title of the notification (optional)
        """
        ep_name = ep_name

        if sickrage.app.config.email_notify_ondownload:
            show = self._parseEp(ep_name)
            to = self._generate_recipients(show)
            if len(to) == 0:
                sickrage.app.log.warning('Skipping email notify because there are no configured recipients')
            else:
                try:
                    msg = MIMEMultipart('alternative')
                    msg.attach(MIMEText(
                        "<body style='font-family:Helvetica, Arial, sans-serif;'><h3>SiCKRAGE Notification - Downloaded</h3>\n<p>Show: <b>" + re.search(
                            "(.+?) -.+", ep_name).group(1) + "</b></p>\n<p>Episode: <b>" + re.search(
                            ".+ - (.+?-.+) -.+", ep_name).group(
                            1) + "</b></p>\n\n<footer style='margin-top: 2.5em; padding: .7em 0; color: #777; border-top: #BBB solid 1px;'>Powered by SiCKRAGE.</footer></body>",
                        'html'))
                except:
                    try:
                        msg = MIMEText(ep_name)
                    except:
                        msg = MIMEText('Episode Downloaded')

                msg['Subject'] = 'Downloaded: ' + ep_name
                msg['From'] = sickrage.app.config.email_from
                msg['To'] = ','.join(to)
                msg['Date'] = formatdate(localtime=True)
                if self._sendmail(sickrage.app.config.email_host, sickrage.app.config.email_port,
                                  sickrage.app.config.email_from,
                                  sickrage.app.config.email_tls,
                                  sickrage.app.config.email_user, sickrage.app.config.email_password, to,
                                  msg):
                    sickrage.app.log.debug("Download notification sent to [%s] for '%s'" % (to, ep_name))
                else:
                    sickrage.app.log.error("Download notification ERROR: %s" % self.last_err)

    def _notify_subtitle_download(self, ep_name, lang, title="Downloaded subtitle:"):
        """
        Send a notification that an subtitle was downloaded

        ep_name: The name of the episode that was downloaded
        lang: Subtitle language wanted
        """
        ep_name = ep_name

        if sickrage.app.config.email_notify_onsubtitledownload:
            show = self._parseEp(ep_name)
            to = self._generate_recipients(show)
            if len(to) == 0:
                sickrage.app.log.warning('Skipping email notify because there are no configured recipients')
            else:
                try:
                    msg = MIMEMultipart('alternative')
                    msg.attach(MIMEText(
                        "<body style='font-family:Helvetica, Arial, sans-serif;'><h3>SiCKRAGE Notification - Subtitle Downloaded</h3>\n<p>Show: <b>" + re.search(
                            "(.+?) -.+", ep_name).group(1) + "</b></p>\n<p>Episode: <b>" + re.search(
                            ".+ - (.+?-.+) -.+", ep_name).group(
                            1) + "</b></p>\n<p>Language: <b>" + lang + "</b></p>\n\n<footer style='margin-top: 2.5em; padding: .7em 0; color: #777; border-top: #BBB solid 1px;'>Powered by SiCKRAGE.</footer></body>",
                        'html'))
                except:
                    try:
                        msg = MIMEText(ep_name + ": " + lang)
                    except:
                        msg = MIMEText("Episode Subtitle Downloaded")

                msg['Subject'] = lang + ' Subtitle Downloaded: ' + ep_name
                msg['From'] = sickrage.app.config.email_from
                msg['To'] = ','.join(to)
                if self._sendmail(sickrage.app.config.email_host, sickrage.app.config.email_port,
                                  sickrage.app.config.email_from,
                                  sickrage.app.config.email_tls,
                                  sickrage.app.config.email_user, sickrage.app.config.email_password, to,
                                  msg):
                    sickrage.app.log.debug("Download notification sent to [%s] for '%s'" % (to, ep_name))
                else:
                    sickrage.app.log.error("Download notification ERROR: %s" % self.last_err)

    def _notify_version_update(self, new_version="??"):
        pass

    def _generate_recipients(self, show):
        addrs = []

        # Grab the global recipients
        for addr in sickrage.app.config.email_list.split(','):
            if (len(addr.strip()) > 0):
                addrs.append(addr)

        # Grab the recipients for the show
        for s in show:
            for subs in [x for x in sickrage.app.showlist if x.name == s]:
                if subs.notify_list:
                    for addr in subs.notify_list.split(','):
                        if len(addr.strip()) > 0:
                            addrs.append(addr)

        addrs = set(addrs)
        sickrage.app.log.debug('Notification recipients: %s' % addrs)
        return addrs

    def _sendmail(self, host, port, smtp_from, use_tls, user, pwd, to, msg, smtpDebug=False):
        sickrage.app.log.debug('HOST: %s; PORT: %s; FROM: %s, TLS: %s, USER: %s, PWD: %s, TO: %s' % (
            host, port, smtp_from, use_tls, user, pwd, to))
        try:
            srv = smtplib.SMTP(host, int(port))
        except Exception as e:
            sickrage.app.log.error("Exception generated while sending e-mail: " + str(e))
            sickrage.app.log.debug(traceback.format_exc())
            return False

        if smtpDebug:
            srv.set_debuglevel(1)
        try:
            if (bool(use_tls)) or (len(user) > 0 and len(pwd) > 0):
                srv.ehlo()
                sickrage.app.log.debug('Sent initial EHLO command!')
            if bool(use_tls):
                srv.starttls()
                sickrage.app.log.debug('Sent STARTTLS command!')
            if len(user) > 0 and len(pwd) > 0:
                srv.login(str(user), str(pwd))
                sickrage.app.log.debug('Sent LOGIN command!')

            srv.sendmail(smtp_from, to, msg.as_string())
            srv.quit()
            return True
        except Exception as e:
            self.last_err = '%s' % e
            return False

    def _parseEp(self, ep_name):
        ep_name = ep_name

        sep = " - "
        titles = ep_name.split(sep)
        titles.sort(key=len, reverse=True)
        sickrage.app.log.debug("TITLES: %s" % titles)
        return titles



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import os
import shutil
import stat

import rarfile

import sickrage
from sickrage.core.common import Quality
from sickrage.core.exceptions import EpisodePostProcessingFailedException, \
    FailedPostProcessingFailedException
from sickrage.core.helpers import is_media_file, is_rar_file, is_hidden_folder, real_path, is_torrent_or_nzb_file, \
    is_sync_file
from sickrage.core.nameparser import InvalidNameException, InvalidShowException, \
    NameParser
from sickrage.core.processors import failed_processor, post_processor


class ProcessResult(object):
    def __init__(self):
        self.result = True
        self.output = ''
        self.missed_files = []
        self.agg_result = True

    def __unicode__(self):
        return self.output


def delete_folder(folder, check_empty=True):
    """
    Removes a folder from the filesystem

    :param folder: Path to folder to remove
    :param check_empty: Boolean, check if the folder is empty before removing it, defaults to True
    :return: True on success, False on failure
    """

    # check if it's a folder
    if not os.path.isdir(folder):
        return False

    # check if it isn't TV_DOWNLOAD_DIR
    if sickrage.app.config.tv_download_dir:
        if real_path(folder) == real_path(sickrage.app.config.tv_download_dir):
            return False

    # check if it's empty folder when wanted checked
    try:
        if check_empty:
            check_files = os.listdir(folder)
            if check_files:
                sickrage.app.log.info(
                    "Not deleting folder {} found the following files: {}".format(folder, check_files))
                return False

            sickrage.app.log.info("Deleting folder (if it's empty): " + folder)
            shutil.rmtree(folder)
        else:
            sickrage.app.log.info("Deleting folder: " + folder)
            shutil.rmtree(folder)
    except (OSError, IOError) as e:
        sickrage.app.log.warning("Warning: unable to delete folder: {}: {}".format(folder, e))
        return False

    return True


def delete_files(processPath, notwantedFiles, result, force=False):
    """
    Remove files from filesystem

    :param processPath: path to process
    :param notwantedFiles: files we do not want
    :param result: Processor results
    :param force: Boolean, force deletion, defaults to false
    """

    if not result.result and force:
        result.output += logHelper("Forcing deletion of files, even though last result was not success",
                                   sickrage.app.log.DEBUG)
    elif not result.result:
        return

    # Delete all file not needed
    for cur_file in notwantedFiles:
        cur_file_path = os.path.join(processPath, cur_file)

        if not os.path.isfile(cur_file_path):
            continue  # Prevent error when a notwantedfiles is an associated files

        result.output += logHelper("Deleting file %s" % cur_file, sickrage.app.log.DEBUG)

        # check first the read-only attribute
        file_attribute = os.stat(cur_file_path)[0]
        if not file_attribute & stat.S_IWRITE:
            # File is read-only, so make it writeable
            result.output += logHelper("Changing ReadOnly Flag for file %s" % cur_file, sickrage.app.log.DEBUG)
            try:
                os.chmod(cur_file_path, stat.S_IWRITE)
            except OSError as e:
                result.output += logHelper(
                    "Cannot change permissions of %s: %s" % (
                        cur_file, str(e.strerror).decode(sickrage.app.sys_encoding)),
                    sickrage.app.log.DEBUG)
        try:
            os.remove(cur_file_path)
        except OSError as e:
            result.output += logHelper(
                "Unable to delete file %s: %s" % (cur_file, str(e.strerror).decode(sickrage.app.sys_encoding)),
                sickrage.app.log.DEBUG)


def logHelper(logMessage, logLevel=None):
    sickrage.app.log.log(logLevel or sickrage.app.log.INFO, logMessage)
    return logMessage + "\n"


def processDir(dirName, nzbName=None, process_method=None, force=False, is_priority=None, delete_on=False, failed=False,
               proc_type="auto", **kwargs):
    """
    Scans through the files in dirName and processes whatever media files it finds

    :param dirName: The folder name to look in
    :param nzbName: The NZB name which resulted in this folder being downloaded
    :param force: True to postprocess already postprocessed files
    :param is_priority: whether to replace the file even if it exists at higher quality
    :param delete_on: delete files and folders after they are processed (always happens with move and auto combination)
    :param failed: Boolean for whether or not the download failed
    :param proc_type: Type of postprocessing auto or manual
    """

    result = ProcessResult()

    # if they passed us a real dir then assume it's the one we want
    if os.path.isdir(dirName):
        dirName = os.path.realpath(dirName)
        result.output += logHelper("Processing in folder {0}".format(dirName), sickrage.app.log.DEBUG)

    # if the client and SickRage are not on the same machine translate the directory into a network directory
    elif all([sickrage.app.config.tv_download_dir,
              os.path.isdir(sickrage.app.config.tv_download_dir),
              os.path.normpath(dirName) == os.path.normpath(sickrage.app.config.tv_download_dir)]):
        dirName = os.path.join(sickrage.app.config.tv_download_dir, os.path.abspath(dirName).split(os.path.sep)[-1])
        result.output += logHelper("Trying to use folder: {0} ".format(dirName), sickrage.app.log.DEBUG)

    # if we didn't find a real dir then quit
    if not os.path.isdir(dirName):
        result.output += logHelper("Unable to figure out what folder to process. "
                                   "If your downloader and SiCKRAGE aren't on the same PC "
                                   "make sure you fill out your TV download dir in the config.",
                                   sickrage.app.log.DEBUG)
        return result

    process_method = process_method or sickrage.app.config.process_method

    directories_from_rars = set()

    # If we have a release name (probably from nzbToMedia), and it is a rar/video, only process that file
    if nzbName and (is_media_file(nzbName) or is_rar_file(nzbName)):
        result.output += logHelper("Processing {}".format(nzbName), sickrage.app.log.INFO)
        generator_to_use = [(dirName, [], [nzbName])]
    else:
        result.output += logHelper("Processing {}".format(dirName), sickrage.app.log.INFO)
        generator_to_use = os.walk(dirName, followlinks=sickrage.app.config.processor_follow_symlinks)

    for current_directory, directory_names, file_names in generator_to_use:
        result.result = True

        file_names = [f for f in file_names if not is_torrent_or_nzb_file(f)]
        rar_files = [x for x in file_names if is_rar_file(os.path.join(current_directory, x))]
        if rar_files:
            extracted_directories = unrar(current_directory, rar_files, force, result)
            if extracted_directories:
                for extracted_directory in extracted_directories:
                    if extracted_directory.split(current_directory)[-1] not in directory_names:
                        result.output += logHelper(
                            "Adding extracted directory to the list of directories to process: {0}".format(
                                extracted_directory), sickrage.app.log.DEBUG
                        )
                        directories_from_rars.add(extracted_directory)

        if not validateDir(current_directory, nzbName, failed, result):
            continue

        video_files = filter(is_media_file, file_names)
        if video_files:
            process_media(current_directory, video_files, nzbName, process_method, force, is_priority, result)
        else:
            result.result = False

        # Delete all file not needed and avoid deleting files if Manual PostProcessing
        if not (process_method == "move" and result.result) or (proc_type == "manual" and not delete_on):
            continue

        # noinspection PyTypeChecker
        unwanted_files = filter(lambda x: x in video_files + rar_files, file_names)
        if unwanted_files:
            result.output += logHelper("Found unwanted files: {0}".format(unwanted_files), sickrage.app.log.DEBUG)

        delete_folder(os.path.join(current_directory, '@eaDir'), False)
        delete_files(current_directory, unwanted_files, result)
        if delete_folder(current_directory, check_empty=not delete_on):
            result.output += logHelper("Deleted folder: {0}".format(current_directory), sickrage.app.log.DEBUG)

    # For processing extracted rars, only allow methods 'move' and 'copy'.
    # On different methods fall back to 'move'.
    method_fallback = ('move', process_method)[process_method in ('move', 'copy')]

    # auto post-processing deletes rar content by default if method is 'move',
    # sickbeard.DELRARCONTENTS allows to override even if method is NOT 'move'
    # manual post-processing will only delete when prompted by delete_on
    delete_rar_contents = any([sickrage.app.config.delrarcontents and proc_type != 'manual',
                               not sickrage.app.config.delrarcontents and proc_type == 'auto' and method_fallback == 'move',
                               proc_type == 'manual' and delete_on])

    for directory_from_rar in directories_from_rars:
        processDir(
            dirName=directory_from_rar,
            nzbName=os.path.basename(directory_from_rar),
            process_method=method_fallback,
            force=force,
            is_priority=is_priority,
            delete_on=delete_rar_contents,
            failed=failed,
            proc_type=proc_type
        )

        # Delete rar file only if the extracted dir was successfully processed
        if proc_type == 'auto' and method_fallback == 'move' or proc_type == 'manual' and delete_on:
            this_rar = [rar_file for rar_file in rar_files if
                        os.path.basename(directory_from_rar) == rar_file.rpartition('.')[0]]
            delete_files(dirName, this_rar, result)  # Deletes only if result.result == True

    result.output += logHelper(("Processing Failed", "Successfully processed")[result.agg_result],
                               (sickrage.app.log.WARNING, sickrage.app.log.INFO)[result.agg_result])
    if result.missed_files:
        result.output += logHelper("Some items were not processed.")
        for missed_file in result.missed_files:
            result.output += logHelper(missed_file)

    return result


def validateDir(process_path, release_name, failed, result):
    """
    Check if directory is valid for processing

    :param process_path: Directory to check
    :param release_name: Original NZB/Torrent name
    :param failed: Previously failed objects
    :param result: Previous results
    :return: True if dir is valid for processing, False if not
    """

    result.output += logHelper("Processing folder " + process_path, sickrage.app.log.DEBUG)

    upper_name = os.path.basename(process_path).upper()
    if upper_name.startswith('_FAILED_') or upper_name.endswith('_FAILED_'):
        result.output += logHelper("The directory name indicates it failed to extract.", sickrage.app.log.DEBUG)
        failed = True
    elif upper_name.startswith('_UNDERSIZED_') or upper_name.endswith('_UNDERSIZED_'):
        result.output += logHelper("The directory name indicates that it was previously rejected for being undersized.",
                                   sickrage.app.log.DEBUG)
        failed = True
    elif upper_name.startswith('_UNPACK') or upper_name.endswith('_UNPACK'):
        result.output += logHelper(
            "The directory name indicates that this release is in the process of being unpacked.",
            sickrage.app.log.DEBUG)
        result.missed_files.append("{0} : Being unpacked".format(process_path))
        return False

    if failed:
        process_failed(process_path, release_name, result)
        result.missed_files.append("{0} : Failed download".format(process_path))
        return False

    if sickrage.app.config.tv_download_dir and real_path(process_path) != real_path(
            sickrage.app.config.tv_download_dir) and is_hidden_folder(process_path):
        result.output += logHelper("Ignoring hidden folder: {0}".format(process_path), sickrage.app.log.DEBUG)
        result.missed_files.append("{0} : Hidden folder".format(process_path))
        return False

    # make sure the dir isn't inside a show dir
    for show in sickrage.app.showlist:
        if process_path.lower().startswith(os.path.realpath(show.location).lower() + os.sep) or \
                process_path.lower() == os.path.realpath(show.location).lower():
            result.output += logHelper(
                "Cannot process an episode that's already been moved to its show dir, skipping " + process_path,
                sickrage.app.log.WARNING)
            return False

    for current_directory, directory_names, file_names in os.walk(process_path, topdown=False,
                                                                  followlinks=sickrage.app.config.processor_follow_symlinks):
        sync_files = filter(is_sync_file, file_names)
        if sync_files and sickrage.app.config.postpone_if_sync_files:
            result.output += logHelper("Found temporary sync files: {0} in path: {1}".format(sync_files,
                                                                                             os.path.join(process_path,
                                                                                                          sync_files[
                                                                                                              0])))
            result.output += logHelper("Skipping post processing for folder: {0}".format(process_path))
            result.missed_files.append("{0} : Sync files found".format(os.path.join(process_path, sync_files[0])))
            continue

        found_files = filter(is_media_file, file_names)
        if sickrage.app.config.unpack == 1:
            found_files += filter(is_rar_file, file_names)

        if current_directory != sickrage.app.config.tv_download_dir and found_files:
            found_files.append(os.path.basename(current_directory))

        for found_file in found_files:
            try:
                NameParser().parse(found_file, cache_result=False)
            except (InvalidNameException, InvalidShowException) as e:
                pass
            else:
                return True

    result.output += logHelper("{0} : No processable items found in folder".format(process_path),
                               sickrage.app.log.DEBUG)
    return False


def unrar(path, rar_files, force, result):
    """
    Extracts RAR files

    :param path: Path to look for files in
    :param rar_files: Names of RAR files
    :param force: process currently processing items
    :param result: Previous results
    :return: List of unpacked file names
    """

    unpacked_dirs = []

    if sickrage.app.config.unpack == 1 and rar_files:
        result.output += logHelper("Packed Releases detected: {0}".format(rar_files), sickrage.app.log.DEBUG)
        for archive in rar_files:
            failure = None
            rar_handle = None
            try:
                archive_path = os.path.join(path, archive)
                if already_postprocessed(path, archive, force, result):
                    result.output += logHelper("Archive file already post-processed, extraction skipped: {}".format
                                               (archive_path), sickrage.app.log.DEBUG)
                    continue

                if not is_rar_file(archive_path):
                    continue

                result.output += logHelper("Checking if archive is valid and contains a video: {}".format(archive_path),
                                           sickrage.app.log.DEBUG)
                rar_handle = rarfile.RarFile(archive_path)
                if rar_handle.needs_password():
                    # TODO: Add support in settings for a list of passwords to try here with rar_handle.set_password(x)
                    result.output += logHelper('Archive needs a password, skipping: {0}'.format(archive_path))
                    continue

                # If there are no video files in the rar, don't extract it
                rar_media_files = filter(is_media_file, rar_handle.namelist())
                if not rar_media_files:
                    continue

                rar_release_name = archive.rpartition('.')[0]

                # Choose the directory we'll unpack to:
                if sickrage.app.config.unpack_dir and os.path.isdir(sickrage.app.config.unpack_dir):
                    unpack_base_dir = sickrage.app.config.unpack_dir
                else:
                    unpack_base_dir = path
                    if sickrage.app.config.unpack_dir:  # Let user know if we can't unpack there
                        result.output += logHelper('Unpack directory cannot be verified. Using {}'.format(path),
                                                   sickrage.app.log.DEBUG)

                # Fix up the list for checking if already processed
                rar_media_files = [os.path.join(unpack_base_dir, rar_release_name, rar_media_file) for rar_media_file in
                                   rar_media_files]

                skip_rar = False
                for rar_media_file in rar_media_files:
                    check_path, check_file = os.path.split(rar_media_file)
                    if already_postprocessed(check_path, check_file, force, result):
                        result.output += logHelper(
                            "Archive file already post-processed, extraction skipped: {0}".format
                            (rar_media_file), sickrage.app.log.DEBUG)
                        skip_rar = True
                        break

                if skip_rar:
                    continue

                rar_extract_path = os.path.join(unpack_base_dir, rar_release_name)
                result.output += logHelper("Unpacking archive: {0}".format(archive), sickrage.app.log.DEBUG)
                rar_handle.extractall(path=rar_extract_path)
                unpacked_dirs.append(rar_extract_path)

            except rarfile.RarCRCError:
                failure = ('Archive Broken', 'Unpacking failed because of a CRC error')
            except rarfile.RarWrongPassword:
                failure = ('Incorrect RAR Password', 'Unpacking failed because of an Incorrect Rar Password')
            except rarfile.PasswordRequired:
                failure = ('Rar is password protected', 'Unpacking failed because it needs a password')
            except rarfile.RarOpenError:
                failure = ('Rar Open Error, check the parent folder and destination file permissions.',
                           'Unpacking failed with a File Open Error (file permissions?)')
            except rarfile.RarExecError:
                failure = ('Invalid Rar Archive Usage',
                           'Unpacking Failed with Invalid Rar Archive Usage. Is unrar installed and on the system PATH?')
            except rarfile.BadRarFile:
                failure = ('Invalid Rar Archive', 'Unpacking Failed with an Invalid Rar Archive Error')
            except rarfile.NeedFirstVolume:
                continue
            except (Exception, rarfile.Error) as e:
                failure = (e, 'Unpacking failed')
            finally:
                if rar_handle:
                    del rar_handle

            if failure:
                result.output += logHelper('Failed to extract the archive {}: {}'.format(archive, failure[0]),
                                           sickrage.app.log.WARNING)
                result.missed_files.append('{} : Unpacking failed: {}'.format(archive, failure[1]))
                result.result = False
                continue

    return unpacked_dirs


def already_postprocessed(dirName, videofile, force, result):
    """
    Check if we already post processed a file

    :param dirName: Directory a file resides in
    :param videofile: File name
    :param force: Force checking when already checking (currently unused)
    :param result: True if file is already postprocessed, False if not
    :return:
    """
    if force:
        return False

    # Avoid processing the same dir again if we use a process method <> move
    if [x for x in sickrage.app.main_db.all('tv_episodes') if
        x['release_name'] and (x['release_name'] in dirName or x['release_name'] in videofile)]:
        return True

    # Needed if we have downloaded the same episode @ different quality
    # But we need to make sure we check the history of the episode we're going to PP, and not others
    np = NameParser(dirName)
    try:
        parse_result = np.parse(dirName)
    except:
        parse_result = False

    for h in (h for h in sickrage.app.main_db.all('history') if h['resource'].endswith(videofile)):
        for e in (e for e in sickrage.app.main_db.get_many('tv_episodes', h['showid'])
                  if h['season'] == e['season'] and h['episode'] == e['episode']
                     and e['status'] in Quality.DOWNLOADED):

            # If we find a showid, a season number, and one or more episode numbers then we need to use those in the query
            if parse_result and (parse_result.indexerid and
                                 parse_result.episode_numbers and
                                 parse_result.season_number):
                if e['showid'] == int(parse_result.indexerid) and \
                        e['season'] == int(parse_result.season_number and
                                           e['episode']) == int(parse_result.episode_numbers[0]):
                    return True
            else:
                return True

    return False


def process_media(processPath, videoFiles, nzbName, process_method, force, is_priority, result):
    """
    Postprocess mediafiles

    :param processPath: Path to postprocess in
    :param videoFiles: Filenames to look for and postprocess
    :param nzbName: Name of NZB file related
    :param process_method: auto/manual
    :param force: Postprocess currently postprocessing file
    :param is_priority: Boolean, is this a priority download
    :param result: Previous results
    """

    processor = None
    for cur_video_file in videoFiles:
        cur_video_file_path = os.path.join(processPath, cur_video_file)

        if already_postprocessed(processPath, cur_video_file, force, result):
            result.output += logHelper("Skipping already processed file: {0}".format(cur_video_file),
                                       sickrage.app.log.DEBUG)
            continue

        try:
            processor = post_processor.PostProcessor(cur_video_file_path, nzbName, process_method, is_priority)
            result.result = processor.process
            process_fail_message = ""
        except EpisodePostProcessingFailedException as e:
            result.result = False
            process_fail_message = e.message

        if processor:
            result.output += processor.log

        if result.result:
            result.output += logHelper("Processing succeeded for " + cur_video_file_path)
        else:
            result.output += logHelper(
                "Processing failed for {0}: {1}".format(cur_video_file_path, process_fail_message),
                sickrage.app.log.WARNING)
            result.missed_files.append("{0} : Processing failed: {1}".format(cur_video_file_path, process_fail_message))
            result.agg_result = False


def process_failed(dirName, nzbName, result):
    """Process a download that did not complete correctly"""

    try:
        processor = failed_processor.FailedProcessor(dirName, nzbName)
        result.result = processor.process()
        process_fail_message = ""
    except FailedPostProcessingFailedException as e:
        processor = None
        result.result = False
        process_fail_message = e

    if processor:
        result.output += processor.log

    if sickrage.app.config.delete_failed and result.result:
        if delete_folder(dirName, check_empty=False):
            result.output += logHelper("Deleted folder: " + dirName, sickrage.app.log.DEBUG)

    if result.result:
        result.output += logHelper("Failed Download Processing succeeded: (" + str(nzbName) + ", " + dirName + ")")
    else:
        result.output += logHelper("Failed Download Processing failed: ({}, {}): {}"
                                   .format(nzbName, dirName, process_fail_message),
                                   sickrage.app.log.WARNING)



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import datetime
import io
import os
from xml.etree.ElementTree import Element, ElementTree, SubElement

import sickrage
from mediabrowser import MediaBrowserMetadata
from sickrage.core.common import dateFormat
from sickrage.core.exceptions import ShowNotFoundException
from sickrage.core.helpers import replaceExtension, indentXML, chmodAsParent
from sickrage.indexers import IndexerApi
from sickrage.indexers.exceptions import indexer_episodenotfound, \
    indexer_error, indexer_seasonnotfound, indexer_shownotfound


class Mede8erMetadata(MediaBrowserMetadata):
    """
    Metadata generation class for Mede8er based on the MediaBrowser.

    The following file structure is used:

    show_root/series.xml                    (show metadata)
    show_root/folder.jpg                    (poster)
    show_root/fanart.jpg                    (fanart)
    show_root/Season ##/folder.jpg          (season thumb)
    show_root/Season ##/filename.ext        (*)
    show_root/Season ##/filename.xml        (episode metadata)
    show_root/Season ##/filename.jpg        (episode thumb)
    """

    def __init__(self,
                 show_metadata=False,
                 episode_metadata=False,
                 fanart=False,
                 poster=False,
                 banner=False,
                 episode_thumbnails=False,
                 season_posters=False,
                 season_banners=False,
                 season_all_poster=False,
                 season_all_banner=False):

        MediaBrowserMetadata.__init__(
            self, show_metadata, episode_metadata, fanart,
            poster, banner, episode_thumbnails, season_posters,
            season_banners, season_all_poster, season_all_banner
        )

        self.name = "Mede8er"

        self.fanart_name = "fanart.jpg"

        # web-ui metadata template
        # self.eg_show_metadata = "series.xml"
        self.eg_episode_metadata = "Season##\\<i>filename</i>.xml"
        self.eg_fanart = "fanart.jpg"
        # self.eg_poster = "folder.jpg"
        # self.eg_banner = "banner.jpg"
        self.eg_episode_thumbnails = "Season##\\<i>filename</i>.jpg"
        # self.eg_season_posters = "Season##\\folder.jpg"
        # self.eg_season_banners = "Season##\\banner.jpg"
        # self.eg_season_all_poster = "<i>not supported</i>"
        # self.eg_season_all_banner = "<i>not supported</i>"

    def get_episode_file_path(self, ep_obj):
        return replaceExtension(ep_obj.location, self._ep_nfo_extension)

    @staticmethod
    def get_episode_thumb_path(ep_obj):
        return replaceExtension(ep_obj.location, 'jpg')

    def _show_data(self, show_obj):
        """
        Creates an elementTree XML structure for a MediaBrowser-style series.xml
        returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """

        indexer_lang = show_obj.lang or sickrage.app.config.indexer_default_language
        lINDEXER_API_PARMS = IndexerApi(show_obj.indexer).api_params.copy()

        lINDEXER_API_PARMS['language'] = indexer_lang

        if show_obj.dvdorder != 0:
            lINDEXER_API_PARMS['dvdorder'] = True

        t = IndexerApi(show_obj.indexer).indexer(**lINDEXER_API_PARMS)

        rootNode = Element("details")
        tv_node = SubElement(rootNode, "movie")
        tv_node.attrib["isExtra"] = "false"
        tv_node.attrib["isSet"] = "false"
        tv_node.attrib["isTV"] = "true"

        try:
            myShow = t[int(show_obj.indexerid)]
        except indexer_shownotfound:
            sickrage.app.log.error(
                "Unable to find show with id " + str(show_obj.indexerid) + " on tvdb, skipping it")
            raise

        except indexer_error:
            sickrage.app.log.error("TVDB is down, can't use its data to make the NFO")
            raise

        # check for title and id
        if not (getattr(myShow, 'seriesname', None) and getattr(myShow, 'id', None)):
            sickrage.app.log.info(
                "Incomplete info for show with id " + str(show_obj.indexerid) + " on " + IndexerApi(
                    show_obj.indexer).name + ", skipping it")
            return False

        SeriesName = SubElement(tv_node, "title")
        SeriesName.text = myShow['seriesname']

        if getattr(myShow, "genre", None):
            Genres = SubElement(tv_node, "genres")
            for genre in myShow['genre'].split('|'):
                if genre and genre.strip():
                    cur_genre = SubElement(Genres, "Genre")
                    cur_genre.text = genre.strip()

        if getattr(myShow, 'firstaired', None):
            FirstAired = SubElement(tv_node, "premiered")
            FirstAired.text = myShow['firstaired']

        if getattr(myShow, "firstaired", None):
            try:
                year_text = str(datetime.datetime.strptime(myShow["firstaired"], dateFormat).year)
                if year_text:
                    year = SubElement(tv_node, "year")
                    year.text = year_text
            except Exception:
                pass

        if getattr(myShow, 'overview', None):
            plot = SubElement(tv_node, "plot")
            plot.text = myShow["overview"]

        if getattr(myShow, 'rating', None):
            try:
                rating = int(float(myShow['rating']) * 10)
            except ValueError:
                rating = 0

            if rating:
                Rating = SubElement(tv_node, "rating")
                Rating.text = str(rating)

        if getattr(myShow, 'status', None):
            Status = SubElement(tv_node, "status")
            Status.text = myShow['status']

        if getattr(myShow, "contentrating", None):
            mpaa = SubElement(tv_node, "mpaa")
            mpaa.text = myShow["contentrating"]

        if getattr(myShow, 'imdb_id', None):
            imdb_id = SubElement(tv_node, "id")
            imdb_id.attrib["moviedb"] = "imdb"
            imdb_id.text = myShow['imdb_id']

        if getattr(myShow, 'id', None):
            indexerid = SubElement(tv_node, "indexerid")
            indexerid.text = myShow['id']

        if getattr(myShow, 'runtime', None):
            Runtime = SubElement(tv_node, "runtime")
            Runtime.text = myShow['runtime']

        cast = SubElement(tv_node, "cast")
        for actor in t.actors(int(show_obj.indexerid)):
            if 'name' in actor and actor['name'].strip():
                cur_actor = SubElement(cast, "actor")
                cur_actor.text = actor['name'].strip()

        indentXML(rootNode)

        data = ElementTree(rootNode)

        return data

    def _ep_data(self, ep_obj):
        """
        Creates an elementTree XML structure for a MediaBrowser style episode.xml
        and returns the resulting data object.

        show_obj: a TVShow instance to create the NFO for
        """

        eps_to_write = [ep_obj] + ep_obj.relatedEps

        indexer_lang = ep_obj.show.lang or sickrage.app.config.indexer_default_language

        try:
            # There's gotta be a better way of doing this but we don't wanna
            # change the language value elsewhere
            lINDEXER_API_PARMS = IndexerApi(ep_obj.show.indexer).api_params.copy()

            lINDEXER_API_PARMS['language'] = indexer_lang

            if ep_obj.show.dvdorder != 0:
                lINDEXER_API_PARMS['dvdorder'] = True

            t = IndexerApi(ep_obj.show.indexer).indexer(**lINDEXER_API_PARMS)
            myShow = t[ep_obj.show.indexerid]
        except indexer_shownotfound as e:
            raise ShowNotFoundException(e.message)
        except indexer_error as e:
            sickrage.app.log.error(
                "Unable to connect to TVDB while creating meta files - skipping - {}".format(e.message))
            return False

        rootNode = Element("details")
        movie = SubElement(rootNode, "movie")

        movie.attrib["isExtra"] = "false"
        movie.attrib["isSet"] = "false"
        movie.attrib["isTV"] = "true"

        # write an MediaBrowser XML containing info for all matching episodes
        for curEpToWrite in eps_to_write:

            try:
                myEp = myShow[curEpToWrite.season][curEpToWrite.episode]
            except (indexer_episodenotfound, indexer_seasonnotfound):
                sickrage.app.log.info(
                    "Unable to find episode %dx%d on %s... has it been removed? Should I delete from db?" %
                    (curEpToWrite.season, curEpToWrite.episode, IndexerApi(ep_obj.show.indexer).name))
                return None

            if curEpToWrite == ep_obj:
                # root (or single) episode

                # default to today's date for specials if firstaired is not set
                if curEpToWrite.season == 0 and not getattr(myEp, 'firstaired', None):
                    myEp['firstaired'] = str(datetime.date.fromordinal(1))

                if not (getattr(myEp, 'episodename', None) and getattr(myEp, 'firstaired', None)):
                    return None

                episode = movie

                if curEpToWrite.name:
                    EpisodeName = SubElement(episode, "title")
                    EpisodeName.text = curEpToWrite.name

                SeasonNumber = SubElement(episode, "season")
                SeasonNumber.text = str(curEpToWrite.season)

                EpisodeNumber = SubElement(episode, "episode")
                EpisodeNumber.text = str(curEpToWrite.episode)

                if getattr(myShow, "firstaired", None):
                    try:
                        year_text = str(datetime.datetime.strptime(myShow["firstaired"], dateFormat).year)
                        if year_text:
                            year = SubElement(episode, "year")
                            year.text = year_text
                    except:
                        pass

                if getattr(myShow, "overview", None):
                    plot = SubElement(episode, "plot")
                    plot.text = myShow["overview"]

                if curEpToWrite.description:
                    Overview = SubElement(episode, "episodeplot")
                    Overview.text = curEpToWrite.description

                if getattr(myShow, 'contentrating', None):
                    mpaa = SubElement(episode, "mpaa")
                    mpaa.text = myShow["contentrating"]

                if not ep_obj.relatedEps and getattr(myEp, "rating", None):
                    try:
                        rating = int((float(myEp['rating']) * 10))
                    except ValueError:
                        rating = 0

                    if rating:
                        Rating = SubElement(episode, "rating")
                        Rating.text = str(rating)

                if getattr(myEp, 'director', None):
                    director = SubElement(episode, "director")
                    director.text = myEp['director']

                if getattr(myEp, 'writer', None):
                    writer = SubElement(episode, "credits")
                    writer.text = myEp['writer']

                cast = SubElement(episode, "cast")
                if getattr(myEp, 'gueststars', None) and isinstance(myEp['gueststars'], basestring):
                    for actor in (x.strip() for x in myEp['gueststars'].split('|') if x.strip()):
                        cur_actor = SubElement(cast, "actor")
                        cur_actor.text = actor

                for actor in t.actors(int(ep_obj.show.indexerid)):
                    if 'name' in actor and actor['name'].strip():
                        cur_actor = SubElement(cast, "actor")
                        cur_actor.text = actor['name'].strip()

            else:
                # append data from (if any) related episodes

                if curEpToWrite.name:
                    if not EpisodeName.text:
                        EpisodeName.text = curEpToWrite.name
                    else:
                        EpisodeName.text = EpisodeName.text + ", " + curEpToWrite.name

                if curEpToWrite.description:
                    if not Overview.text:
                        Overview.text = curEpToWrite.description
                    else:
                        Overview.text = Overview.text + "\r" + curEpToWrite.description

        indentXML(rootNode)
        data = ElementTree(rootNode)

        return data

    def write_show_file(self, show_obj):
        """
        Generates and writes show_obj's metadata under the given path to the
        filename given by get_show_file_path()

        show_obj: TVShow object for which to create the metadata

        path: An absolute or relative path where we should put the file. Note that
                the file name will be the default show_file_name.

        Note that this method expects that _show_data will return an ElementTree
        object. If your _show_data returns data in another format yo'll need to
        override this method.
        """

        data = self._show_data(show_obj)

        if not data:
            return False

        nfo_file_path = self.get_show_file_path(show_obj)
        nfo_file_dir = os.path.dirname(nfo_file_path)

        try:
            if not os.path.isdir(nfo_file_dir):
                sickrage.app.log.debug("Metadata dir didn't exist, creating it at " + nfo_file_dir)
                os.makedirs(nfo_file_dir)
                chmodAsParent(nfo_file_dir)

            sickrage.app.log.debug("Writing show nfo file to " + nfo_file_path)

            nfo_file = io.open(nfo_file_path, 'wb')

            data.write(nfo_file)
            nfo_file.close()
            chmodAsParent(nfo_file_path)
        except IOError as e:
            sickrage.app.log.error(
                "Unable to write file to " + nfo_file_path + " - are you sure the folder is writable? {}".format(
                    e.message))
            return False

        return True

    def write_ep_file(self, ep_obj):
        """
        Generates and writes ep_obj's metadata under the given path with the
        given filename root. Uses the episode's name with the extension in
        _ep_nfo_extension.

        ep_obj: TVEpisode object for which to create the metadata

        file_name_path: The file name to use for this metadata. Note that the extension
                will be automatically added based on _ep_nfo_extension. This should
                include an absolute path.

        Note that this method expects that _ep_data will return an ElementTree
        object. If your _ep_data returns data in another format yo'll need to
        override this method.
        """

        data = self._ep_data(ep_obj)

        if not data:
            return False

        nfo_file_path = self.get_episode_file_path(ep_obj)
        nfo_file_dir = os.path.dirname(nfo_file_path)

        try:
            if not os.path.isdir(nfo_file_dir):
                sickrage.app.log.debug("Metadata dir didn't exist, creating it at " + nfo_file_dir)
                os.makedirs(nfo_file_dir)
                chmodAsParent(nfo_file_dir)

            sickrage.app.log.debug("Writing episode nfo file to " + nfo_file_path)

            nfo_file = io.open(nfo_file_path, 'wb')

            data.write(nfo_file)
            nfo_file.close()
            chmodAsParent(nfo_file_path)
        except IOError as e:
            sickrage.app.log.error(
                "Unable to write file to " + nfo_file_path + " - are you sure the folder is writable? {}".format(
                    e.message))
            return False

        return True



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import io
import mimetypes
import os

from .exceptions import NoParserError

PARSERS = [('asf', ['video/asf'], ['asf', 'wmv', 'wma']),
           ('flv', ['video/flv'], ['flv']),
           ('mkv', ['video/x-matroska', 'application/mkv'], ['mkv', 'mka', 'webm']),
           ('mp4', ['video/quicktime', 'video/mp4'], ['mov', 'qt', 'mp4', 'mp4a', '3gp', '3gp2', '3g2', 'mk2']),
           ('mpeg', ['video/mpeg'], ['mpeg', 'mpg', 'mp4', 'ts']),
           ('ogm', ['application/ogg'], ['ogm', 'ogg', 'ogv']),
           ('real', ['video/real'], ['rm', 'ra', 'ram']),
           ('riff', ['video/avi'], ['wav', 'avi'])
           ]


def parse(path):
    """Parse metadata of the given video

    :param string path: path to the video file to parse
    :return: a parser corresponding to the video's mimetype or extension
    :rtype: :class:`~enzyme.core.AVContainer`

    """
    if not os.path.isfile(path):
        raise ValueError('Invalid path')
    extension = os.path.splitext(path)[1][1:]
    mimetype = mimetypes.guess_type(path)[0]
    parser_ext = None
    parser_mime = None
    for (parser_name, parser_mimetypes, parser_extensions) in PARSERS:
        if mimetype in parser_mimetypes:
            parser_mime = parser_name
        if extension in parser_extensions:
            parser_ext = parser_name
    parser = parser_mime or parser_ext
    if not parser:
        raise NoParserError()
    mod = __import__(parser, globals=globals(), locals=locals(), fromlist=[], level=-1)
    with io.open(path, 'rb') as f:
        p = mod.Parser(f)
    return p
